#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
"""This file contains the Record class for metaknowledge and one helper function for parsing WOS records, recordParser. The record class is used to represent a single records meta-data from WOS.
"""
import itertools
import io
import collections

from ..mkRecord import ExtendedRecord

from ..WOS.tagProcessing.funcDicts import tagNameConverterDict, tagToFull
from ..WOS.tagProcessing.tagFunctions import tagToFunc
from ..mkExceptions import BadWOSFile, BadWOSRecord

class WOSRecord(ExtendedRecord):
    """Class for full WOS records

    It is meant to be immutable; many of the methods and attributes are evaluated when first called, not when the object is created, and the results are stored privately.

    The record's meta-data is stored in an ordered dictionary labeled by WOS tags. To access the raw data stored in the original record the [tags()](./CollectionWithIDs.html#metaknowledge.CollectionWithIDs.tags) method can be used. To access data that has been processed and cleaned the attributes named after the tags are used.

    # Customizations

    The `Record`'s hashing and equality testing are based on the WOS number (the tag is 'UT', and also called the accession number). They are strings starting with `'WOS:'` and followed by 15 or so numbers and letters, although both the length and character set are known to vary. The numbers are unique to each record so are used for comparisons. If a record is `bad`  all equality checks return `False`.

    When converted to a string the records title is used so for a record `R`, `R.TI == R.title == str(R)` and its representation uses the WOS number instead of memory location.

    # Attributes

    When a record is created if the parsing of the WOS file failed it is marked as `bad`. The `bad` attribute is set to True and the `error` attribute is created to contain the exception object.

    Generally, to get the information from a Record its attributes should be used. For a Record `R`, calling `R.CR` causes [citations()](../modules/WOS.html#metaknowledge.WOS.tagProcessing.tagFunctions.citations) from the the [tagProcessing](../modules/WOS.html#module-metaknowledge.WOS.tagProcessing.helpFuncs) module to be called on the contents of the raw 'CR' field. Then the result is saved and returned. In this case, a list of Citation objects is returned. You can also call `R.citations` to get the same effect, as each known field tag has a longer name (currently there are 61 field tags). These names are meant to make accessing tags more readable and mapping from tag to name can be found in the tagToFull dict. If a tag is known (in [tagToFull](../modules/WOS.html#metaknowledge.WOS.tagProcessing.funcDicts.tagToFull)) but not in the raw data `None` is returned instead. Most tags when cleaned return a string or list of strings, the exact results can be found in the help for the particular function.

    The attribute `authors` is also defined as a convenience and returns the same as 'AF' or if that is not found 'AU'.

    # \_\_Init\_\_

    Records are generally created as collections in  [Recordcollections](./RecordCollection.html#metaknowledge.RecordCollection), and not as individual objects. If you wish to create one on its own it is possible, the arguments are as follows.

    # Parameters

    _inRecord_: `files stream, dict, str or itertools.chain`

    > If it is a file stream the file must be open at the location of the first tag in the record, usually 'PT', and the file will be read until 'ER' is found, which indicates the end of the record in the file.

    > If a dict is passed the dictionary is used as the database of fields and tags, so each key is considered a WOS tag and each value a list of the lines of the original associated with the tag. This is the same form of dict that [recordParser](../modules/WOS.html#metaknowledge.WOS.recordWOS.recordParser) returns.

    > For a string the input must be the raw textual data of a single record in the WOS style, like the file stream it must start at the first tag and end in `'ER'`.

    > itertools.chain is treated identically to a file stream and is used by [RecordCollections](./RecordCollection.html#metaknowledge.RecordCollection).

    _sFile_ : `optional [str]`

    > Is the name of the file the raw data was in, by default it is blank. It is mostly used to make error messages more informative.

    _sLine_ : `optional [int]`

    > Is the line the record starts on in the raw data file. It is mostly used to make error messages more informative.
    """

    def __init__(self, inRecord, sFile = "", sLine = 0):
        """See help on [Record](./Record.html#metaknowledge.Record) for details"""
        bad = False
        error = None
        fieldDict = None
        try:
            if isinstance(inRecord, dict) or isinstance(inRecord, collections.OrderedDict):
                fieldDict = collections.OrderedDict(inRecord)
            elif isinstance(inRecord, itertools.chain):
                fieldDict = recordParser(inRecord)
            elif isinstance(inRecord, io.IOBase):
                fieldDict = recordParser(enumerate(inRecord))
            elif isinstance(inRecord, str):
                def addChartoEnd(lst):
                    for s in lst:
                        yield s + '\n'
                fieldDict = recordParser(enumerate(addChartoEnd(inRecord.split('\n')), start = 1))
                #string io
            else:
                raise TypeError("Unsupported input type '{}', WOSRecords cannot be created from '{}'".format(inRecord, type(inRecord)))
        except BadWOSRecord as b:
            self.bad = True
            self.error = b
            fieldDict = collections.OrderedDict()
        if fieldDict is not None:
            if 'UT' in fieldDict:
                self._wosNum = fieldDict['UT'][0]
            else:
                self._wosNum = 'WOS:Missing'
                bad = True
                error = BadWOSRecord("Missing WOS number")
        ExtendedRecord.__init__(self, fieldDict, self._wosNum, bad, error, sFile = sFile, sLine = sLine)

    def encoding(self):
        return 'utf-8'

    @staticmethod
    def getAltName(tag):
        return tagNameConverterDict.get(tag)

    @staticmethod
    def tagProcessingFunc(tag):
        return tagToFunc.get(tag, lambda x: x)

    def specialFuncs(self, key):
        if key == 'selfCitation':
            #this cannot be computed until after the Record is created
            return self.createCitation()
        elif key == 'id':
            return self.id
        elif key == 'grants':
            return []
        else:
            raise KeyError("WOS Records do not have any special functions besides 'grants', they are what the special functions are mostly based on.")

    @property
    def wosString(self):
        """Returns the WOS number (UT tag) of the record"""
        return self._wosNum

    @property
    def UT(self):
        """Returns the UT tag (WOS number) of the record"""
        return self._wosNum

    def writeRecord(self, infile):
        """Writes to _infile_ the original contents of the Record. This is intended for use by [RecordCollections](./RecordCollection.html#metaknowledge.RecordCollection) to write to file. What is written to _infile_ is bit for bit identical to the original record file (if utf-8 is used). No newline is inserted above the write but the last character is a newline.

        # Parameters

        _infile_ : `file stream`

        > An open utf-8 encoded file
        """
        if self.bad:
            raise BadWOSRecord("This record cannot be converted to a file as the input was malformed.\nThe original line number (if any) is: {} and the original file is: '{}'".format(self._sourceLine, self._sourceFile))
        else:
            for tag in self._fieldDict.keys():
                for i, value in enumerate(self._fieldDict[tag]):
                    if i == 0:
                        infile.write(tag + ' ')
                    else:
                        infile.write('   ')
                    infile.write(value + '\n')
            infile.write("ER\n")

def recordParser(paper):
    """This is function that is used to create [Records](../classes/Record.html#metaknowledge.Record) from files.

    **recordParser**() reads the file _paper_ until it reaches 'ER'. For each field tag it adds an entry to the returned dict with the tag as the key and a list of the entries as the value, the list has each line separately, so for the following two lines in a record:

        AF BREVIK, I
           ANICIN, B

    The entry in the returned dict would be `{'AF' : ["BREVIK, I", "ANICIN, B"]}`

    `Record` objects can be created with these dictionaries as the initializer.

    # Parameters

    _paper_ : `file stream`

    > An open file, with the current line at the beginning of the WOS record.

    # Returns

    `OrderedDict[str : List[str]]`

    > A dictionary mapping WOS tags to lists, the lists are of strings, each string is a line of the record associated with the tag.
    """
    tagList = []
    doneReading = False
    l = (0, '')
    for l in paper:
        if len(l[1]) < 3:
            #Line too short
            raise BadWOSRecord("Missing field on line {} : {}".format(l[0], l[1]))
        elif 'ER' in l[1][:2]:
            #Reached the end of the record
            doneReading = True
            break
        elif l[1][2] != ' ':
            #Field tag longer than 2 or offset in some way
            raise BadWOSFile("Field tag not formed correctly on line " + str(l[0]) + " : " + l[1])
        elif '   ' in l[1][:3]: #the string is three spaces in row
            #No new tag append line to current tag (last tag in tagList)
            tagList[-1][1].append(l[1][3:-1])
        else:
            #New tag create new entry at the end of tagList
            tagList.append((l[1][:2], [l[1][3:-1]]))
    if not doneReading:
        raise BadWOSRecord("End of file reached before ER: {}".format(l[1]))
    else:
        retdict = collections.OrderedDict(tagList)
        if len(retdict) == len(tagList):
            return retdict
        else:
            dupSet = set()
            for tupl in tagList:
                if tupl[0] in retdict:
                    dupSet.add(tupl[0])
            raise BadWOSRecord("Duplicate tags (" + ', '.join(dupSet) + ") in record")



proQuestSpecialTagToFunc = {
    #'DOI' : lambda R : None,
    'authorsShort' : lambda R : R['Author'],
    #'grants' : lambda R : [],
    #'address' : '',
    'selfCitation' : lambda R: R.createCitation(),
    'beginningPage' : lambda R : 0,
    'keywords' : lambda R : R['Identifier / keyword'],
    'abstract' : lambda R : R['Abstract'],
    #'citations' : '',
    'authorsFull' : lambda R : 'Author',
    #'volume' : '',
    'year' : lambda R : R['Degree date'],
    #'j9' : '',
    #'journal' : '',
    'title' : lambda R : R['Title'],
    'id' : lambda R : R.id,
}



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
import metaknowledge

if __name__ == "__main__":
    rlst = metaknowledge.wosParser("testFile.isi")
    s = '['
    for R in rlst:
        s +=(str(R.__getstate__()) + ',\n')
    s += ']'
    print(s)



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
import networkx as nx
import csv
import os

from .progressBar import _ProgressBar
from .mkExceptions import RCValueError

import metaknowledge

def readGraph(edgeList, nodeList = None, directed = False, idKey = 'ID', eSource = 'Source', eDest = 'Target'):
    """Reads the files given by _edgeList_ and _nodeList_ and creates a networkx graph for the files.

    This is designed only for the files produced by metaknowledge and is meant to be the reverse of [writeGraph()](#metaknowledge.graphHelpers.writeGraph), if this does not produce the desired results the networkx builtin [networkx.read_edgelist()](https://networkx.github.io/documentation/networkx-1.10/reference/generated/networkx.readwrite.edgelist.read_edgelist.html) could be tried as it is aimed at a more general usage.

    The read edge list format assumes the column named _eSource_ (default `'Source'`) is the source node, then the column _eDest_ (default `'Target'`) givens the destination and all other columns are attributes of the edges, e.g. weight.

    The read node list format assumes the column _idKey_ (default `'ID'`) is the ID of the node for the edge list and the resulting network. All other columns are considered attributes of the node, e.g. count.

    **Note**: If the names of the columns do not match those given to **readGraph()** a `KeyError` exception will be raised.

    **Note**: If nodes appear in the edgelist but not the nodeList they will be created silently with no attributes.

    # Parameters

    _edgeList_ : `str`

    > a string giving the path to the edge list file

    _nodeList_ : `optional [str]`

    > default `None`, a string giving the path to the node list file

    _directed_ : `optional [bool]`

    > default `False`, if `True` the produced network is directed from _eSource_ to _eDest_

    _idKey_ : `optional [str]`

    > default `'ID'`, the name of the ID column in the node list

    _eSource_ : `optional [str]`

    > default `'Source'`, the name of the source column in the edge list

    _eDest_ : `optional [str]`

    > default `'Target'`, the name of the destination column in the edge list

    # Returns

    `networkx Graph`

    > the graph described by the input files
    """
    progArgs = (0, "Starting to reading graphs")
    if metaknowledge.VERBOSE_MODE:
        progKwargs = {'dummy' : False}
    else:
        progKwargs = {'dummy' : True}
    with _ProgressBar(*progArgs, **progKwargs) as PBar:
        if directed:
            grph = nx.DiGraph()
        else:
            grph = nx.Graph()
        if nodeList:
            PBar.updateVal(0, "Reading " + nodeList)
            f = open(os.path.expanduser(os.path.abspath(nodeList)))
            nFile = csv.DictReader(f)
            for line in nFile:
                vals = line
                ndID = vals[idKey]
                del vals[idKey]
                if len(vals) > 0:
                    grph.add_node(ndID, **vals)
                else:
                    grph.add_node(ndID)
            f.close()
        PBar.updateVal(.25, "Reading " + edgeList)
        f = open(os.path.expanduser(os.path.abspath(edgeList)))
        eFile = csv.DictReader(f)
        for line in eFile:
            vals = line
            eFrom = vals[eSource]
            eTo = vals[eDest]
            del vals[eSource]
            del vals[eDest]
            if len(vals) > 0:
                grph.add_edge(eFrom, eTo, **vals)
            else:
                grph.add_edge(eFrom, eTo)
        PBar.finish("{} nodes and {} edges found".format(len(grph.nodes()), len(grph.edges())))
        f.close()
        return grph

def writeGraph(grph, name, edgeInfo = True, typing = False, suffix = 'csv', overwrite = True, allSameAttribute = False):
    """Writes both the edge list and the node attribute list of _grph_ to files starting with _name_.

    The output files start with _name_, the file type (edgeList, nodeAttributes) then if typing is True the type of graph (directed or undirected) then the suffix, the default is as follows:

    >> name_fileType.suffix

    Both files are csv's with comma delimiters and double quote quoting characters. The edge list has two columns for the source and destination of the edge, `'Source'` and `'Target'` respectively, then, if _edgeInfo_ is `True`, for each attribute of the node another column is created. The node list has one column call "ID" with the node ids used by networkx and all other columns are the node attributes.

    To read back these files use [readGraph()](#metaknowledge.graphHelpers.readGraph) and to write only one type of lsit use [writeEdgeList()](#metaknowledge.graphHelpers.writeEdgeList) or [writeNodeAttributeFile()](#metaknowledge.graphHelpers.writeNodeAttributeFile).

    **Warning**: this function will overwrite files, if they are in the way of the output, to prevent this set _overwrite_ to `False`

    **Note**: If any nodes or edges are missing an attribute a `KeyError` will be raised.

    # Parameters

    _grph_ : `networkx Graph`

    > A networkx graph of the network to be written.

    _name_ : `str`

    > The start of the file name to be written, can include a path.

    _edgeInfo_ : `optional [bool]`

    > Default `True`, if `True` the the attributes of each edge are written to the edge list.

    _typing_ : `optional [bool]`

    > Default `False`, if `True` the directed ness of the graph will be added to the file names.

    _suffix_ : `optional [str]`

    > Default `"csv"`, the suffix of the file.

    _overwrite_ : `optional [bool]`

    > Default `True`, if `True` files will be overwritten silently, otherwise an `OSError` exception will be raised.
    """
    progArgs = (0, "Writing the graph to files starting with: {}".format(name))
    if metaknowledge.VERBOSE_MODE:
        progKwargs = {'dummy' : False}
    else:
        progKwargs = {'dummy' : True}
    with _ProgressBar(*progArgs, **progKwargs) as PBar:
        if typing:
            if isinstance(grph, nx.classes.digraph.DiGraph) or isinstance(grph, nx.classes.multidigraph.MultiDiGraph):
                grphType = "_directed"
            else:
                grphType = "_undirected"
        else:
            grphType = ''
        nameCompts = os.path.split(os.path.expanduser(os.path.normpath(name)))
        if nameCompts[0] == '' and nameCompts[1] == '':
            edgeListName = "edgeList"+ grphType + '.' + suffix
            nodesAtrName = "nodeAttributes"+ grphType + '.' + suffix
        elif nameCompts[0] == '':
            edgeListName = nameCompts[1] + "_edgeList"+ grphType + '.' + suffix
            nodesAtrName = nameCompts[1] + "_nodeAttributes"+ grphType + '.' + suffix
        elif nameCompts[1] == '':
            edgeListName = os.path.join(nameCompts[0], "edgeList"+ grphType + '.' + suffix)
            nodesAtrName = os.path.join(nameCompts[0], "nodeAttributes"+ grphType + '.' + suffix)
        else:
            edgeListName = os.path.join(nameCompts[0], nameCompts[1] + "_edgeList"+ grphType + '.' + suffix)
            nodesAtrName = os.path.join(nameCompts[0], nameCompts[1] + "_nodeAttributes"+ grphType + '.' + suffix)
        if not overwrite:
            if os.path.isfile(edgeListName):
                raise OSError(edgeListName+ " already exists")
            if os.path.isfile(nodesAtrName):
                raise OSError(nodesAtrName + " already exists")
        writeEdgeList(grph, edgeListName, extraInfo = edgeInfo, allSameAttribute = allSameAttribute, _progBar = PBar)
        writeNodeAttributeFile(grph, nodesAtrName, allSameAttribute = allSameAttribute, _progBar = PBar)
        PBar.finish("{} nodes and {} edges written to file".format(len(grph.nodes()), len(grph.edges())))

def writeEdgeList(grph, name, extraInfo = True, allSameAttribute = False, _progBar = None):
    """Writes an edge list of _grph_ at the destination _name_.

    The edge list has two columns for the source and destination of the edge, `'Source'` and `'Target'` respectively, then, if _edgeInfo_ is `True`, for each attribute of the node another column is created.

    **Note**: If any edges are missing an attribute it will be left blank by default, enable _allSameAttribute_ to cause a `KeyError` to be raised.

    # Parameters

    _grph_ : `networkx Graph`

    > The graph to be written to _name_

    _name_ : `str`

    > The name of the file to be written

    _edgeInfo_ : `optional [bool]`

    > Default `True`, if `True` the attributes of each edge will be written

    _allSameAttribute_ : `optional [bool]`

    > Default `False`, if `True` all the edges must have the same attributes or an exception will be raised. If `False` the missing attributes will be left blank.
    """
    count = 0
    eMax = len(grph.edges())
    if metaknowledge.VERBOSE_MODE or isinstance(_progBar, _ProgressBar):
        if isinstance(_progBar, _ProgressBar):
            PBar = _progBar
            PBar.updateVal(0, "Writing edge list {}".format(name))
        else:
            PBar = _ProgressBar(0, "Writing edge list {}".format(name))
    else:
        PBar = _ProgressBar(0, "Writing edge list {}".format(name), dummy = True)
    if len(grph.edges(data = True)) < 1:
        outFile = open(os.path.expanduser(os.path.abspath(name)), 'w')
        outFile.write('"Source","Target"\n')
        outFile.close()
        PBar.updateVal(1, "Done edge list '{}', 0 edges written.".format(name))
    else:
        if extraInfo:
            csvHeader = []
            if allSameAttribute:
                csvHeader = ['Source'] +  ['Target'] + list(grph.edges(data = True).__next__()[2].keys())
            else:
                extraAttribs = set()
                for eTuple in grph.edges(data = True):
                    count += 1
                    if count % 1000 == 0:
                        PBar.updateVal(count / eMax * .10, "Checking over edge: '{}' to '{}'".format(eTuple[0], eTuple[1]))
                    s = set(eTuple[2].keys()) - extraAttribs
                    if len(s) > 0:
                        for i in s:
                            extraAttribs.add(i)
                csvHeader = ['Source', 'Target'] + list(extraAttribs)
        else:
            csvHeader = ['Source'] +  ['Target']
        count = 0
        PBar.updateVal(.01, "Opening file {}".format(name))
        f = open(os.path.expanduser(os.path.abspath(name)), 'w', newline = '')
        outFile = csv.DictWriter(f, csvHeader, delimiter = ',', quotechar = '"', quoting=csv.QUOTE_NONNUMERIC)
        outFile.writeheader()
        if extraInfo:
            for e in grph.edges(data = True):
                count += 1
                if count % 1000 == 0:
                    PBar.updateVal(count / eMax * .90 + .10, "Writing edge: '{}' to '{}'".format(e[0], e[1]))
                eDict = e[2].copy()
                eDict['Source'] = e[0]
                eDict['Target'] = e[1]
                try:
                    outFile.writerow(eDict)
                except UnicodeEncodeError:
                    #Because Windows
                    newDict = {k.encode('ASCII', errors='ignore').decode('ASCII', errors='ignore') if isinstance(k, str) else k: v.encode('ASCII', errors='ignore').decode('ASCII', errors='ignore') if isinstance(v, str) else v for k, v in eDict.items()}
                    outFile.writerow(newDict)
                except ValueError:
                    raise ValueError("Some edges in The graph do not have the same attributes")
        else:
            for e in grph.edges():
                count += 1
                if count % 1000 == 0:
                    PBar.updateVal(count / eMax * .90 + .10, "Writing edge: '{}' to '{}'".format(e[0], e[1]))
                eDict['Source'] = e[0]
                eDict['Target'] = e[1]
                try:
                    outFile.writerow(eDict)
                except UnicodeEncodeError:
                    #Because Windows
                    newDict = {k.encode('ASCII', errors='ignore').decode('ASCII', errors='ignore') if isinstance(k, str) else k: v.encode('ASCII', errors='ignore').decode('ASCII', errors='ignore') if isinstance(v, str) else v for k, v in eDict.items()}
                    outFile.writerow(newDict)
        PBar.updateVal(1, "Closing {}".format(name))
        f.close()
        if not isinstance(_progBar, _ProgressBar):
            PBar.finish("Done edge list {}, {} edges written.".format(name, count))

def writeNodeAttributeFile(grph, name, allSameAttribute = False, _progBar = None):
    """Writes a node attribute list of _grph_ to the file given by the path _name_.

    The node list has one column call `'ID'` with the node ids used by networkx and all other columns are the node attributes.

    **Note**: If any nodes are missing an attribute it will be left blank by default, enable _allSameAttribute_ to cause a `KeyError` to be raised.

    # Parameters

    _grph_ : `networkx Graph`

    > The graph to be written to _name_

    _name_ : `str`

    > The name of the file to be written

    _allSameAttribute_ : `optional [bool]`

    > Default `False`, if `True` all the nodes must have the same attributes or an exception will be raised. If `False` the missing attributes will be left blank.
    """
    count = 0
    nMax = len(grph.nodes())
    if metaknowledge.VERBOSE_MODE or isinstance(_progBar, _ProgressBar):
        if isinstance(_progBar, _ProgressBar):
            PBar = _progBar
            PBar.updateVal(0, "Writing node list {}".format(name))
        else:
            PBar = _ProgressBar(0, "Writing node list {}".format(name))
    else:
        PBar = _ProgressBar(0, "Writing node list {}".format(name), dummy = True)
    if len(grph.nodes(data = True)) < 1:
        outFile = open(os.path.expanduser(os.path.abspath(name)), 'w')
        outFile.write('ID\n')
        outFile.close()
        PBar.updateVal(1, "Done node attribute list: {}, 0 nodes written.".format(name))
    else:
        csvHeader = []
        if allSameAttribute:
            csvHeader = ['ID'] + list(grph.nodes(data = True).__next__()[1].keys())
        else:
            extraAttribs = set()
            for n, attribs in grph.nodes(data = True):
                count += 1
                if count % 100 == 0:
                    PBar.updateVal(count / nMax * .10, "Checking over node: '{}'".format(n))
                s = set(attribs.keys()) - extraAttribs
                if len(s) > 0:
                    for i in s:
                        extraAttribs.add(i)
            csvHeader = ['ID'] + list(extraAttribs)
        count = 0
        PBar.updateVal(.10, "Opening '{}'".format(name))
        f = open(name, 'w', newline = '')
        outFile = csv.DictWriter(f, csvHeader, delimiter = ',', quotechar = '"', quoting = csv.QUOTE_NONNUMERIC)
        outFile.writeheader()
        for n in grph.nodes(data = True):
            count += 1
            if count % 100 == 0:
                PBar.updateVal(count / nMax * .90 + .10, "Writing node: '{}'".format(n[0]))
            nDict = n[1].copy()
            nDict['ID'] = n[0]
            try:
                outFile.writerow(nDict)
            except UnicodeEncodeError:
                #Because Windows
                newDict = {k.encode('ASCII', errors='ignore').decode('ASCII', errors='ignore') if isinstance(k, str) else k: v.encode('ASCII', errors='ignore').decode('ASCII', errors='ignore') if isinstance(v, str) else v for k, v in nDict.items()}
                outFile.writerow(newDict)
            except ValueError:
                raise ValueError("Some nodes in the graph do not have the same attributes")
        PBar.updateVal(1, "Closing {}".format(name))
        f.close()
        if not isinstance(_progBar, _ProgressBar):
            PBar.finish("Done node attribute list: {}, {} nodes written.".format(name, count))

def writeTnetFile(grph, name, modeNameString, weighted = False, sourceMode = None, timeString = None, nodeIndexString = 'tnet-ID', weightString = 'weight'):
    """Writes an edge list designed for reading by the _R_ package [tnet](https://toreopsahl.com/tnet/).

    The _networkx_ graph provided must be a pure two-mode network, the modes must be 2 different values for the node attribute accessed by _modeNameString_ and all edges must be between different node types. Each node will be given an integer id, stored in the attribute given by _nodeIndexString_, these ids are then written to the file as the endpoints of the edges. Unless _sourceMode_ is given which mode is the source (first column) and which the target (second column) is random.

    **Note** the _grph_ will be modified by this function, the ids of the nodes will be written to the graph at the attribute _nodeIndexString_.

    # Parameters

    _grph_ : `network Graph`

    > The graph that will be written to _name_

    _name_ : `str`

    > The path of the file to write

    _modeNameString_ : `str`

    > The name of the attribute _grph_'s modes are stored in

    _weighted_ : `optional bool`

    > Default `False`, if `True` then the attribute _weightString_ will be written to the weight column

    _sourceMode_ : `optional str`

    > Default `None`, if given the name of the mode used for the source (first column) in the output file

    _timeString_ : `optional str`

    > Default `None`, if present the attribute _timeString_ of an edge will be written to the time column surrounded by double quotes (").

    **Note** The format used by tnet for dates is very strict it uses the ISO format, down to the second and without time zones.

    _nodeIndexString_ : `optional str`

    > Default `'tnet-ID'`, the name of the attribute to save the id for each node

    _weightString_ : `optional str`

    > Default `'weight'`, the name of the weight attribute
    """
    count = 0
    eMax = len(grph.edges())
    progArgs = (0, "Writing tnet edge list {}".format(name))
    if metaknowledge.VERBOSE_MODE:
        progKwargs = {'dummy' : False}
    else:
        progKwargs = {'dummy' : True}
    with _ProgressBar(*progArgs, **progKwargs) as PBar:
        if sourceMode is not None:
            modes = [sourceMode]
        else:
            modes = []
        mode1Set = set()
        PBar.updateVal(.1, "Indexing nodes for tnet")
        for nodeIndex, node in enumerate(grph.nodes(data = True), start = 1):
            try:
                nMode = node[1][modeNameString]
            except KeyError:
                #too many modes so will fail
                modes = [1,2,3]
                nMode = 4
            if nMode not in modes:
                if len(modes) < 2:
                    modes.append(nMode)
                else:
                    raise RCValueError("Too many modes of '{}' found in the network or one of the nodes was missing its mode. There must be exactly 2 modes.".format(modeNameString))
            if nMode == modes[0]:
                mode1Set.add(node[0])
            node[1][nodeIndexString] = nodeIndex
        if len(modes) != 2:
            raise RCValueError("Too few modes of '{}' found in the network. There must be exactly 2 modes.".format(modeNameString))
        with open(name, 'w', encoding = 'utf-8') as f:
            edgesCaller = {'data' : True}
            if timeString is not None:
                edgesCaller['keys'] = True
            for *nodes, eDict in grph.edges(**edgesCaller):
                if timeString is not None:
                    n1, n2, keyVal = nodes
                else:
                    n1, n2 = nodes
                count += 1
                if count % 1000 == 1:
                    PBar.updateVal(count/ eMax * .9 + .1, "writing edge: '{}'-'{}'".format(n1, n2))
                if n1 in mode1Set:
                    if n2 in mode1Set:
                        raise RCValueError("The nodes '{}' and '{}' have an edge and the same type. The network must be purely 2-mode.".format(n1, n2))
                elif n2 in mode1Set:
                    n1, n2 = n2, n1
                else:
                    raise RCValueError("The nodes '{}' and '{}' have an edge and the same type. The network must be purely 2-mode.".format(n1, n2))
                if timeString is not None:
                    eTimeString = '"{}" '.format(keyVal)
                else:
                    eTimeString = ''
                if weighted:
                    f.write("{}{} {} {}\n".format(eTimeString, grph.node[n1][nodeIndexString], grph.node[n2][nodeIndexString], eDict[weightString]))
                else:
                    f.write("{}{} {}\n".format(eTimeString, grph.node[n1][nodeIndexString], grph.node[n2][nodeIndexString]))
        PBar.finish("Done writing tnet file '{}'".format(name))

def getWeight(grph, nd1, nd2, weightString = "weight", returnType = int):
    """
    A way of getting the weight of an edge with or without weight as a parameter
    returns a the value of the weight parameter converted to returnType if it is given or 1 (also converted) if not
    """
    if not weightString:
        return returnType(1)
    else:
        return returnType(grph.edges[nd1, nd2][weightString])

def getNodeDegrees(grph, weightString = "weight", strictMode = False,  returnType = int, edgeType = 'bi'):
    """
    Retunrs a dictionary of nodes to their degrees, the degree is determined by adding the weight of edge with the weight being the string weightString that gives the name of the attribute of each edge containng thier weight. The Weights are then converted to the type returnType. If weightString is give as False instead each edge is counted as 1.

    edgeType, takes in one of three strings: 'bi', 'in', 'out'. 'bi' means both nodes on the edge count it, 'out' mans only the one the edge comes form counts it and 'in' means only the node the edge goes to counts it. 'bi' is the default. Use only on directional graphs as otherwise the selected nodes is random.
    """
    ndsDict = {}
    for nd in grph.nodes():
        ndsDict[nd] = returnType(0)
    for e in grph.edges(data = True):
        if weightString:
            try:
                edgVal = returnType(e[2][weightString])
            except KeyError:
                if strictMode:
                    raise KeyError("The edge from " + str(e[0]) + " to " + str(e[1]) + " does not have the attribute: '" + str(weightString) + "'")
                else:
                    edgVal = returnType(1)
        else:
            edgVal = returnType(1)
        if edgeType == 'bi':
            ndsDict[e[0]] += edgVal
            ndsDict[e[1]] += edgVal
        elif edgeType == 'in':
            ndsDict[e[1]] += edgVal
        elif edgeType == 'out':
            ndsDict[e[0]] += edgVal
        else:
            raise ValueError("edgeType must be 'bi', 'in', or 'out'")
    return ndsDict

def getDegreeDistribution(grph, weightParameter = "weight", strictWeightNames = False,  weightType = int, directionalType = 'bi'):
    if weightType != int:
        raise ValueError("Unsupported type for weights, only ints are supported")
    ndsDict = getNodeDegrees(grph, weightString = weightParameter, strictMode = strictWeightNames, returnType = weightType, edgeType = directionalType)
    distVec = [0] * (max(ndsDict.values()) + 1)
    for v in ndsDict.values():
        distVec[v] += 1
    return distVec

def dropEdges(grph, minWeight = - float('inf'), maxWeight = float('inf'), parameterName = 'weight', ignoreUnweighted = False, dropSelfLoops = False):
    """Modifies _grph_ by dropping edges whose weight is not within the inclusive bounds of _minWeight_ and _maxWeight_, i.e after running _grph_ will only have edges whose weights meet the following inequality: _minWeight_ <= edge's weight <= _maxWeight_. A `Keyerror` will be raised if the graph is unweighted unless _ignoreUnweighted_ is `True`, the weight is determined by examining the attribute _parameterName_.

    **Note**: none of the default options will result in _grph_ being modified so only specify the relevant ones, e.g. `dropEdges(G, dropSelfLoops = True)` will remove only the self loops from `G`.

    # Parameters

    _grph_ : `networkx Graph`

    > The graph to be modified.

    _minWeight_ : `optional [int or double]`

    > default `-inf`, the minimum weight for an edge to be kept in the graph.

    _maxWeight_ : `optional [int or double]`

    > default `inf`, the maximum weight for an edge to be kept in the graph.

    _parameterName_ : `optional [str]`

    > default `'weight'`, key to weight field in the edge's attribute dictionary, the default is the same as networkx and metaknowledge so is likely to be correct

    _ignoreUnweighted_ : `optional [bool]`

    > default `False`, if `True` unweighted edges will kept

    _dropSelfLoops_ : `optional [bool]`

    > default `False`, if `True` self loops will be removed regardless of their weight
    """
    count = 0
    total = len(grph.edges())
    if metaknowledge.VERBOSE_MODE:
        progArgs = (0, "Dropping edges")
        progKwargs = {}
    else:
        progArgs = (0, "Dropping edges")
        progKwargs = {'dummy' : True}
    with _ProgressBar(*progArgs, **progKwargs) as PBar:
        if dropSelfLoops:
            slps = list(nx.selfloop_edges(grph))

            PBar.updateVal(0, "Dropping self {} loops".format(len(slps)))
            for e in slps:
                grph.remove_edge(e[0], e[1])
        edgesToDrop = []
        if minWeight != - float('inf') or maxWeight != float('inf'):
            for e in grph.edges(data = True):
                try:
                    val = e[2][parameterName]
                except KeyError:
                    if not ignoreUnweighted:
                        raise KeyError("One or more Edges do not have weight or " + str(parameterName), " is not the name of the weight")
                    else:
                        pass
                else:

                    count += 1
                    if count % 100000 == 0:
                        PBar.updateVal(count/ total, str(count) + " edges analysed and " + str(total -len(grph.edges())) + " edges dropped")
                    if val > maxWeight or  val < minWeight:
                        edgesToDrop.append((e[0], e[1]))
        grph.remove_edges_from(edgesToDrop)
        PBar.finish(str(total - len(grph.edges())) + " edges out of " + str(total) + " dropped, " + str(len(grph.edges())) + " returned")

def dropNodesByDegree(grph, minDegree = -float('inf'), maxDegree = float('inf'), useWeight = True, parameterName = 'weight', includeUnweighted = True):
    """Modifies _grph_ by dropping nodes that do not have a degree that is within inclusive bounds of _minDegree_ and _maxDegree_, i.e after running _grph_ will only have nodes whose degrees meet the following inequality: _minDegree_ <= node's degree <= _maxDegree_.

    Degree is determined in two ways, the default _useWeight_ is the weight attribute of the edges to a node will be summed, the attribute's name is _parameterName_ otherwise the number of edges touching the node is used. If _includeUnweighted_ is `True` then _useWeight_ will assign a degree of 1 to unweighted edges.


    # Parameters

    _grph_ : `networkx Graph`

    > The graph to be modified.

    _minDegree_ : `optional [int or double]`

    > default `-inf`, the minimum degree for an node to be kept in the graph.

    _maxDegree_ : `optional [int or double]`

    > default `inf`, the maximum degree for an node to be kept in the graph.

    _useWeight_ : `optional [bool]`

    > default `True`, if `True` the the edge weights will be summed to get the degree, if `False` the number of edges will be used to determine the degree.

    _parameterName_ : `optional [str]`

    > default `'weight'`, key to weight field in the edge's attribute dictionary, the default is the same as networkx and metaknowledge so is likely to be correct.

    _includeUnweighted_ : `optional [bool]`

    > default `True`, if `True` edges with no weight will be considered to have a weight of 1, if `False` they will cause a `KeyError` to be raised.
    """
    count = 0
    total = len(grph.nodes())
    if metaknowledge.VERBOSE_MODE:
        progArgs = (0, "Dropping nodes by degree")
        progKwargs = {}
    else:
        progArgs = (0, "Dropping nodes by degree")
        progKwargs = {'dummy' : True}
    with _ProgressBar(*progArgs, **progKwargs) as PBar:
        badNodes = []
        for n in grph.nodes():
            if PBar:
                count += 1
                if count % 10000 == 0:
                    PBar.updateVal(count/ total, str(count) + " nodes analysed and " + str(len(badNodes)) + " nodes dropped")
            val = 0
            if useWeight:
                for e in grph.edges(n, data = True):
                    try:
                        val += e[2][parameterName]
                    except KeyError:
                        if not includeUnweighted:
                            raise KeyError("One or more Edges do not have weight or " + str(parameterName), " is not the name of the weight")
                        else:
                            val += 1
            else:
                val = len(grph.edges(n))
            if val < minDegree or val > maxDegree:
                badNodes.append(n)
        if PBar:
            PBar.updateVal(1, "Cleaning up graph")
        grph.remove_nodes_from(badNodes)
        if PBar:
            PBar.finish("{} nodes out of {} dropped, {} returned".format(len(badNodes), total, total - len(badNodes)))


def dropNodesByCount(grph, minCount = -float('inf'), maxCount = float('inf'), parameterName = 'count', ignoreMissing = False):
    """Modifies _grph_ by dropping nodes that do not have a count that is within inclusive bounds of _minCount_ and _maxCount_, i.e after running _grph_ will only have nodes whose degrees meet the following inequality: _minCount_ <= node's degree <= _maxCount_.

    Count is determined by the count attribute, _parameterName_, and if missing will result in a `KeyError` being raised. _ignoreMissing_ can be set to `True` to suppress the error.

    minCount and maxCount default to negative and positive infinity respectively so without specifying either the output should be the input

    # Parameters

    _grph_ : `networkx Graph`

    > The graph to be modified.

    _minCount_ : `optional [int or double]`

    > default `-inf`, the minimum Count for an node to be kept in the graph.

    _maxCount_ : `optional [int or double]`

    > default `inf`, the maximum Count for an node to be kept in the graph.

    _parameterName_ : `optional [str]`

    > default `'count'`, key to count field in the nodes's attribute dictionary, the default is the same thoughout metaknowledge so is likely to be correct.

    _ignoreMissing_ : `optional [bool]`

    > default `False`, if `True` nodes missing a count will be kept in the graph instead of raising an exception
    """
    count = 0
    total = len(grph.nodes())
    if metaknowledge.VERBOSE_MODE:
        progArgs = (0, "Dropping nodes by count")
        progKwargs = {}
    else:
        progArgs = (0, "Dropping nodes by count")
        progKwargs = {'dummy' : True}
    with _ProgressBar(*progArgs, **progKwargs) as PBar:
        badNodes = []
        for n in grph.nodes(data = True):
            if PBar:
                count += 1
                if count % 10000 == 0:
                    PBar.updateVal(count/ total, str(count) + "nodes analysed and {} nodes dropped".format(len(badNodes)))
            try:
                val = n[1][parameterName]
            except KeyError:
                if not ignoreMissing:
                    raise KeyError("One or more nodes do not have counts or " + str(parameterName), " is not the name of the count parameter")
                else:
                    pass
            else:
                if val < minCount or val > maxCount:
                    badNodes.append(n[0])
        if PBar:
            PBar.updateVal(1, "Cleaning up graph")
        grph.remove_nodes_from(badNodes)
        if PBar:
            PBar.finish("{} nodes out of {} dropped, {} returned".format(len(badNodes), total, total - len(badNodes)))

def mergeGraphs(targetGraph, addedGraph, incrementedNodeVal = 'count', incrementedEdgeVal = 'weight'):
    """A quick way of merging graphs, this is meant to be quick and is only intended for graphs generated by metaknowledge. This does not check anything and as such may cause unexpected results if the source and target were not generated by the same method.

    **mergeGraphs**() will **modify** _targetGraph_ in place by adding the nodes and edges found in the second, _addedGraph_. If a node or edge exists _targetGraph_ is given precedence, but the edge and node attributes given by _incrementedNodeVal_ and incrementedEdgeVal are added instead of being overwritten.

    # Parameters

    _targetGraph_ : `networkx Graph`

    > the graph to be modified, it has precedence.

    _addedGraph_ : `networkx Graph`

    > the graph that is unmodified, it is added and does **not** have precedence.

    _incrementedNodeVal_ : `optional [str]`

    > default `'count'`, the name of the count attribute for the graph's nodes. When merging this attribute will be the sum of the values in the input graphs, instead of _targetGraph_'s value.

    _incrementedEdgeVal_ : `optional [str]`

    > default `'weight'`, the name of the weight attribute for the graph's edges. When merging this attribute will be the sum of the values in the input graphs, instead of _targetGraph_'s value.
    """

    for addedNode, attribs in addedGraph.nodes(data = True):
        if incrementedNodeVal:
            try:
                targetGraph.node[addedNode][incrementedNodeVal] += attribs[incrementedNodeVal]
            except KeyError:
                targetGraph.add_node(addedNode, **attribs)
        else:
            if not targetGraph.has_node(addedNode):
                targetGraph.add_node(addedNode, **attribs)
    for edgeNode1, edgeNode2, attribs in addedGraph.edges(data = True):
        if incrementedEdgeVal:
            try:
                targetGraph.edges[edgeNode1, edgeNode2][incrementedEdgeVal] += attribs[incrementedEdgeVal]
            except KeyError:
                targetGraph.add_edge(edgeNode1, edgeNode2, **attribs)
        else:
            if not targetGraph.Graph.has_edge(edgeNode1, edgeNode2):
                targetGraph.add_edge(edgeNode1, edgeNode2, **attribs)

def graphStats(G, stats = ('nodes', 'edges', 'isolates', 'loops', 'density', 'transitivity'), makeString = True, sentenceString = False):
    """Returns a string or list containing statistics about the graph _G_.

    **graphStats()** gives 6 different statistics: number of nodes, number of edges, number of isolates, number of loops, density and transitivity. The ones wanted can be given to _stats_. By default a string giving each stat on a different line it can also produce a sentence containing all the requested statistics or the raw values can be accessed instead by setting _makeString_ to `False`.

    # Parameters

    _G_ : `networkx Graph`

    > The graph for the statistics to be determined of

    _stats_ : `optional [list or tuple [str]]`

    > Default `('nodes', 'edges', 'isolates', 'loops', 'density', 'transitivity')`, a list or tuple containing any number or combination of the strings:

    > `"nodes"`, `"edges"`, `"isolates"`, `"loops"`, `"density"` and `"transitivity"``

    > At least one occurrence of the corresponding string causes the statistics to be provided in the string output. For the non-string (tuple) output the returned tuple has the same length as the input and each output is at the same index as the string that requested it, e.g.

    > `_stats_ = ("edges", "loops", "edges")`

    > The return is a tuple with 2 elements the first and last of which are the number of edges and the second is the number of loops

    _makeString_ : `optional [bool]`

    > Default `True`, if `True` a string is returned if `False` a tuple

    _sentenceString_ : `optional [bool]`

    >Default `False` : if `True` the returned string is a sentce, otherwise each value has a seperate line.

    # Returns

    `str or tuple [float and int]`

    > The type is determined by _makeString_ and the layout by _stats_
     """

    for sts in stats:
        if sts not in ['nodes', 'edges', 'isolates', 'loops', 'density', 'transitivity']:
            raise RuntimeError('"{}" is not a valid stat.'.format(sts))
    if makeString:
        stsData = []
    else:
        stsData = {}
    if 'nodes' in stats:
        if makeString:
            if sentenceString:
                stsData.append("{:G} nodes".format(len(G.nodes())))
            else:
                stsData.append("Nodes: {:G}".format(len(G.nodes())))
        else:
            stsData['nodes'] = len(G.nodes())
    if 'edges' in stats:
        if makeString:
            if sentenceString:
                stsData.append("{:G} edges".format(len(G.edges())))
            else:
                stsData.append("Edges: {:G}".format(len(G.edges())))
        else:
            stsData['edges'] = len(G.edges())
    if 'isolates' in stats:
        if makeString:
            if sentenceString:
                stsData.append("{:G} isolates".format(len(list(nx.isolates(G)))))
            else:
                stsData.append("Isolates: {:G}".format(len(list(nx.isolates(G)))))
        else:
            stsData['isolates'] = len(list(nx.isolates(G)))
    if 'loops' in stats:
        if makeString:
            if sentenceString:
                stsData.append("{:G} self loops".format(len(list(nx.selfloop_edges(G)))))
            else:
                stsData.append("Self loops: {:G}".format(len(list(nx.selfloop_edges(G)))))
        else:
            stsData['loops'] = len(list(nx.selfloop_edges(G)))
    if 'density' in stats:
        if makeString:
            if sentenceString:
                stsData.append("a density of {:G}".format(nx.density(G)))
            else:
                stsData.append("Density: {:G}".format(nx.density(G)))
        else:
            stsData['density'] = nx.density(G)
    if 'transitivity' in stats:
        if makeString:
            if sentenceString:
                stsData.append("a transitivity of {:G}".format(nx.transitivity(G)))
            else:
                stsData.append("Transitivity: {:G}".format(nx.transitivity(G)))
        else:
            stsData['transitivity'] = nx.transitivity(G)
    if makeString:
        if sentenceString:
            retString = "The graph has "
            if len(stsData) < 1:
                return retString
            elif len(stsData) == 1:
                return retString + stsData[0]
            else:
                return retString + ', '.join(stsData[:-1]) + ' and ' + stsData[-1]
        else:
            return '\n'.join(stsData)
    else:
        retLst = []
        for sts in stats:
            retLst.append(stsData[sts])
        return tuple(retLst)



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2016
import unittest
import metaknowledge

import os

class TestScopus(unittest.TestCase):

    def setUp(self):
        metaknowledge.VERBOSE_MODE = False
        self.RC = metaknowledge.RecordCollection("metaknowledge/tests/scopus_testing.csv.scopus")
        self.R = self.RC.peek()

    def test_creation(self):
        Rstart = self.RC.peek()
        R = metaknowledge.ScopusRecord(Rstart._fieldDict)
        self.assertEqual(R, Rstart)
        with open("metaknowledge/tests/scopus_testing.csv.scopus") as f:
            f.read(1)
            header = f.readline()[:-1].split(',')
            R = metaknowledge.ScopusRecord(f.readline(), header = header)
            self.assertEqual(R.id, 'EID:2-s2.0-84963944162')
            R = metaknowledge.ScopusRecord(f.readline(), header = header)
            self.assertEqual(R.id, 'EID:2-s2.0-84943362392')
        with self.assertRaises(TypeError):
            R = metaknowledge.ScopusRecord(12345678)
        R = metaknowledge.ScopusRecord(",2132,4,3fdgf,fgdgdfdg,dgfdg,,,,,,,,,,,,,,,,,,,2e5r6t789765432\n")
        self.assertTrue(R.bad)
        with self.assertRaises(metaknowledge.BadScopusRecord):
            R.writeRecord('not a file')

    def test_isCollection(self):
        self.assertIsInstance(self.RC, metaknowledge.RecordCollection)

    def test_isScopus(self):
        self.assertIsInstance(self.R, metaknowledge.ScopusRecord)

    def test_specials(self):
        for R in self.RC:
            for s in metaknowledge.scopus.scopusSpecialTagToFunc.keys():
                self.assertIsInstance(R.get(s), (str, type(None), list, int, metaknowledge.Citation))

    def test_allFields(self):
        for R in self.RC:
            for k,v in R.items():
                self.assertIsInstance(k, str)
                self.assertIsInstance(v, (str, list, int))

    def test_graphs(self):
        self.assertEqual(metaknowledge.graphStats(self.RC.networkCoAuthor(), sentenceString = True), "The graph has 1798 nodes, 89236 edges, 36 isolates, 15 self loops, a density of 0.0552422 and a transitivity of 0.994673")
        self.assertEqual(metaknowledge.graphStats(self.RC.networkCitation(), sentenceString = True), "The graph has 10026 nodes, 10362 edges, 0 isolates, 0 self loops, a density of 0.000103094 and a transitivity of 0")

    def test_write(self):
        fileName = 'tempFile.scopus.tmp'
        self.RC.writeFile(fileName)
        self.assertEqual(os.path.getsize(fileName), os.path.getsize("metaknowledge/tests/scopus_testing.csv.scopus") + 11511) #Not quite identical due to double quotes
        os.remove(fileName)



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
import argparse
import re
import os.path
import subprocess

args = argparse.Namespace()
codeRegex = re.compile(r'\[([0-9]*)\](.*)')

startString = """{
 "cells": [
  """

endString = """  ],
 "metadata": {
   "kernelspec": {
   "display_name": "Python 3",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
 },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.4.3"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 0
}"""

def argumentParser():
    parser = argparse.ArgumentParser(description="A simple script to convert markdown (.md) files to iPython Notebooks (.ipynb)")
    #parser.add_argument("--output", "-o")
    #parser.add_argument("--execute", "-e", action = "store_true", default = False)
    parser.add_argument("files", type=argparse.FileType('r'), default = [], nargs = '*')

    return parser.parse_args()

def convertString(file):
    currentExNum = float('inf')
    currentBufferType = ''
    stringBuffer = ''
    stringResult = []
    for line in file.readlines():
        code = re.match(codeRegex,line)
        if stringBuffer == '':
            if code:
                stringBuffer = code.group(2) + '\n'
                currentExNum = code.group(1)
                currentBufferType = 'py'
            else:
                stringBuffer = line
                currentBufferType = 'md'
                currentExNum = float('inf')
        else:
            if code:
                if currentBufferType == 'py' and currentExNum == code.group(1):
                    stringBuffer += code.group(2) + '\n'
                elif currentBufferType == 'py':
                    stringResult.append(writePYcell(stringBuffer, currentExNum))
                    stringBuffer = code.group(2)+ '\n'
                    currentExNum = code.group(1)
                    currentBufferType = 'py'
                else:
                    stringResult.append(writeMDcell(stringBuffer))
                    stringBuffer = code.group(2)+ '\n'
                    currentExNum = code.group(1)
                    currentBufferType = 'py'
            else:
                if currentBufferType == 'md':
                    if line == '\n' and stringBuffer[-2:] == '\n\n':
                        stringResult.append(writeMDcell(stringBuffer[:-1]))
                        stringBuffer = ''
                        currentBufferType = ''
                    else:
                        stringBuffer += line
                else:
                    stringResult.append(writePYcell(stringBuffer, currentExNum))
                    stringBuffer = line
                    currentBufferType = 'md'
                    currentExNum = float('inf')
    if stringBuffer != '':
        if currentBufferType == 'md':
            stringResult.append(writeMDcell(stringBuffer))
        else:
            stringResult.append(writePYcell(stringBuffer, excount = currentExNum))
    return startString + ',\n  '.join(stringResult) + endString

def convert(file):
    nameCompts = os.path.splitext( os.path.expanduser(os.path.normpath(file.name)))
    fileName = nameCompts[0] + '.ipynb'
    outFile = open(fileName, 'w+')
    outFile.write(convertString(file))
    return fileName


def stringPreprossesing(s):
    s = s.lstrip('\n')
    s = s.replace(r'"', r'\"')
    s = s.replace('\n', '\\n",\n     "')[:-11]
    return s


def writeMDcell(s):
    return """{{
   "cell_type": "markdown",
   "metadata": {{}},
   "source": [
    "{0}"
   ]
  }}""".format(stringPreprossesing(s))

def writePYcell(s, excount = ''):
    return """{{
   "cell_type": "code",
   "execution_count": null,
   "metadata": {{
     "collapsed": false
   }},
   "outputs": [],
   "source": [
    "{}"
   ]
  }}""".format(stringPreprossesing(s))

def mkMdToNb():
    args = argumentParser()
    for f in args.files:
        fname = convert(f)
if __name__ =='__main__':
    mkMdToNb()



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2016

import csv

from .recordScopus import ScopusRecord, scopusHeader

from ..mkExceptions import BadScopusFile

def isScopusFile(infile, checkedLines = 2, maxHeaderDiff = 3):
    """Determines if _infile_ is the path to a Scopus csv file. A file is considerd to be a Scopus file if it has the correct encoding (`utf-8` with BOM (Byte Order Mark)) and within the first _checkedLines_ a line contains the complete header, the list of all header entries in order is found in [`scopus.scopusHeader`](#metaknowledge.scopus).

    **Note** this is for csv files _not_ plain text files from scopus, plain text files are not complete.

    # Parameters

    _infile_ : `str`

    > The path to the targets file

    _checkedLines_ : `optional [int]`

    > default 2, the number of lines to check for the header

    _maxHeaderDiff_ : `optional [int]`

    > default 3, maximum number of different entries in the potetial file from the current known header `metaknowledge.scopus.scopusHeader`, if exceeded an `False` will be returned

    # Returns

    `bool`

    > `True` if the file is a Scopus csv file
    """
    try:
        with open(infile, 'r', encoding='utf-8') as openfile:
            if openfile.read(1) != "\ufeff":
                return False
            for i in range(checkedLines):
                if len(set(openfile.readline()[:-1].split(',')) ^ set(scopusHeader)) < maxHeaderDiff:
                    return True
    except (StopIteration, UnicodeDecodeError):
        return False
    else:
        return False

def scopusParser(scopusFile):
    """Parses a scopus file, _scopusFile_, to extract the individual lines as [ScopusRecords](../classes/ScopusRecord.html#metaknowledge.scopus.ScopusRecord).

    A Scopus file is a csv (Comma-separated values) with a complete header, see [`scopus.scopusHeader`](#metaknowledge.scopus) for the entries, and each line after it containing a record's entry. The string valued entries are quoted with double quotes which means double quotes inside them can cause issues, see [scopusRecordParser()](#metaknowledge.scopus.recordScopus.scopusRecordParser) for more information.

    # Parameters

    _scopusFile_ : `str`

    > A path to a valid scopus file, use [isScopusFile()](#metaknowledge.scopus.scopusHandlers.isScopusFile) to verify

    # Returns

    `set[ScopusRecord]`

    > Records for each of the entries
    """
    #assumes the file is Scopus
    recSet = set()
    error = None
    lineNum = 0
    try:
        with open(scopusFile, 'r', encoding = 'utf-8') as openfile:
            #Get rid of the BOM
            openfile.read(1)
            header = openfile.readline()[:-1].split(',')
            if len(set(header) ^ set(scopusHeader)) == 0:
                header = None
            lineNum = 0
            try:
                for line, row in enumerate(openfile, start = 2):
                    lineNum = line
                    recSet.add(ScopusRecord(row, header = header, sFile = scopusFile, sLine = line))
            except BadScopusFile as e:
                if error is None:
                    error = BadScopusFile("The file '{}' becomes unparsable after line: {}, due to the error: {} ".format(scopusFile, lineNum, e))
    except (csv.Error, UnicodeDecodeError):
        if error is None:
            error = BadScopusFile("The file '{}' has parts of it that are unparsable starting at line: {}.".format(scopusFile, lineNum))
    return recSet, error



class mkException(Exception):
    pass

class CollectionTypeError(mkException, TypeError):
    pass

class RCTypeError(mkException, TypeError):
    pass

class TagError(mkException):
    pass

class RCValueError(mkException):
    pass

class BadInputFile(mkException):
    pass

class BadRecord(mkException):
    pass

class BadPubmedRecord(mkException):
    pass

class BadPubmedFile(mkException):
    pass

class BadScopusRecord(mkException):
    pass

class BadScopusFile(mkException):
    pass

class BadProQuestRecord(mkException):
    pass

class BadProQuestFile(mkException):
    pass

class RecordsNotCompatible(mkException):
    pass

class JournalDataBaseError(mkException):
    pass

class GenderException(mkException):
    pass

class cacheError(mkException):
    """Exception raised when loading a cached RecordCollection fails, should only be seen inside metaknowledge and always be caught."""
    pass

class BadWOSRecord(BadRecord):
    """Exception thrown by the [record parser](../modules/WOS.html#metaknowledge.WOS.recordWOS.recordParser) to indicate a mis-formated record. This occurs when some component of the record does not parse. The messages will be any of:

        * _Missing field on line (line Number):(line)_, which indicates a line was to short, there should have been a tag followed by information

        * _End of file reached before ER_, which indicates the file ended before the 'ER' indicator appeared, 'ER' indicates the end of a record. This is often due to a copy and paste error.

        * _Duplicate tags in record_, which indicates the record had 2 or more lines with the same tag.

        * _Missing WOS number_, which indicates the record did not have a 'UT' tag.

    Records with a BadWOSRecord error are likely incomplete or the combination of two or more single records.
    """
    pass

class BadWOSFile(Warning):
    """Exception thrown by wosParser for mis-formated files
    """
    pass

class BadCitation(Warning):
    """
    Exception thrown by Citation
    """
    pass

class BadGrant(mkException):
    pass


class GrantCollectionException(mkException):
    pass

class UnknownFile(mkException):
    pass



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2016
import unittest
import metaknowledge

class TestGrants(unittest.TestCase):
    def setUp(self):
        self.Grant1 = metaknowledge.MedlineGrant("U10 HD04267/HG/NICHD NHI HHS/Unit State")
        self.Grant2 = metaknowledge.MedlineGrant("HG/NICHD NHI HHS/Unit State")

    def test_isGrant(self):
        self.assertIsInstance(self.Grant1, metaknowledge.Grant)
        self.assertIsInstance(self.Grant2, metaknowledge.Grant)

    def test_init(self):
        Gshort = metaknowledge.MedlineGrant("U10 HD04267/NICHD NHI HHS/Unit State")
        Gmid = metaknowledge.MedlineGrant("U10 /HD04267HG/NICHD NHI HHS/Unit State")
        Glong = metaknowledge.MedlineGrant("U/10 /HD042/67HG/NICHD NHI HHS/Unit State")
        self.assertNotEqual(Gshort, Glong)
        self.assertNotEqual(Gmid, Glong)

    def test_bad(self):
        G = metaknowledge.MedlineGrant("NICHD NHI HHSUnit State")
        self.assertTrue(G.bad)

    def test_eq(self):
        self.assertNotEqual(1, self.Grant2)
        self.assertNotEqual(self.Grant1, self.Grant2)

    def test_hash(self):
        self.assertIsInstance(hash(self.Grant1), int)

    def test_orgin(self):
        self.assertEqual("U10 HD04267/HG/NICHD NHI HHS/Unit State", self.Grant1.original)

    def test_rerp(self):
        self.assertEqual(repr(self.Grant1), "<metaknowledge.MedlineGrant object U10 HD04267/HG-Unit State>")

    def test_NSF(self):
        GC = metaknowledge.GrantCollection("metaknowledge/tests/nsfTestFiles")
        G = GC.networkMultiMode(GC.tags())
        self.assertEqual(metaknowledge.graphStats(G, sentenceString = True), "The graph has 244 nodes, 2077 edges, 0 isolates, 19 self loops, a density of 0.0703974 and a transitivity of 0.497237")



"""`Record` is the base of various objects in mk, it is intended to be used with things that have some sort of key-value relationship and is basiclly a hashable python dict. It also has a few extra attributes intead to make debugging and record keeping easier.

+ `bad` cand be set to `True` to indcate something is wrong with the issue being saved in `error` the exact details are left to designer

+ `_sourceFile` and `_sourceLine` store the original file name and line number and are mostly for improving error messages

+ `_id` should be a unique string, that preferably can be used to identify the record from its source, although the latter is not always possible to do so, do your best. It is also what is used for hashing and comparison

+ `_fieldDict` contains the base mapping of keys to values, it is the dictionary

`ExtendedRecord` is what WOSRecord and its ilk inherit from and extends `Record` by adding memoizing and processing of the fields. `ExtendedRecord` cannot be invoked directly as it has many abstract (virtual) methods that define how the tags are to be proccesed what they are called, what encoding to use when writing to disk, etc.
"""

import abc
try:
    import collections.abc
except ImportError:
    import collections
    collections.abc = collections
import copy

from .constants import commonRecordFields

from .mkExceptions import BadRecord

from .genders.nameGender import recordGenders

class Record(collections.abc.Mapping, collections.abc.Hashable):
    """A dictionary with error handling and an id string.

    `Record` is the base class of the all objects in _metaknowledge_ that contain information as key-value pairs, these are the grants and the records from different sources.

    The error handling of the `Record` is done with the `bad` attribute. If there is some issue with the data _bad_ should be `True` and _error_ given an `Exception` that was caused by or explains the error.

    # Customizations

    `Record` is a subclass of `abc.collections.Mapping` which means it has almost all the methods a dictionary does, the missing ones are those that modify entries. So to access the value of the key `'title'` from a `Record` `R`, you would use either the square brace notation `t = R['title']` or the `get()` function `t = R.get('title')` just like a dictionary. The other methods like `keys()` or `copy()` also work.

    In addition to being a mapping `Records` are also hashable with their hashes being based on a unique id string they are given on creation, usually some kind of accession number the source gives them. The two optional arguments _sFile_ and _sLine_, which should be given the name of the file the records came from and the line it started on respectively, are used to make the errors more useful.

    # \_\_Init\_\_

    _fieldDict_ is the dictionary the `Record` will use and _idValue_ is the unique identifier of the `Record`.

    # Parameters

    _fieldDict_ : `dict[str:]`

    > A dictionary that maps from strings to values

    _idValue_ : `str`

    > A unique identifier string for the `Record`

    _bad_ : `bool`

    > `True` if there are issues with the `Record`, otherwise `False`

    _error_ : `Exception`

    > The `Exception` that caused whatever error made the record be marked as bad or `None`

    _sFile_ : `str`

    > A string that gives the source file of the original records

    _sLine_ : `int`

    > The first line the original record is found on in the source file
    """

    #This is for the documentation generation, it doesn't do anything on its own
    _documented = ['__hash__', '__eq__', '__str__', '__repr__']

    def __init__(self, fieldDict, idValue, bad, error, sFile = "", sLine = 0):
        #File stuff for debug/error messages
        self._sourceFile = sFile
        self._sourceLine = sLine

        #Error message stuff
        self.bad = bad
        self.error = error

        #Important data
        self._id = idValue
        self._fieldDict = fieldDict

    #collections.abc.Hashable method

    def __hash__(self):
        """Gives a hash of the id or if `bad` returns a hash of the fields combined with the error messages, either of these could be blank

        `bad` Records are more likely to cause hash collisions due to their lack of entropy when created.

        # Returns

        `int`

        > A hopefully unique random number
        """
        if self.bad:
            return hash(str(self._fieldDict.values()) + str(self.error))
        else:
            return hash(self._id)

    #collections.abc.Mapping methods

    def __getitem__(self, key):
        """This is redfined as something interesting for ExtendedRecord"""
        try:
            return self._fieldDict[key]
        except KeyError:
            raise KeyError("'{}' could not be found in the Record".format(key))

    def __iter__(self):
        """Iterates over the tags in the Record"""
        for t in self._fieldDict:
            yield t

    def __len__(self):
        """Returns the number of tags"""
        return len(self._fieldDict)

    def __contains__(self, item):
        """Checks if the tag _item_ is in the Record"""
        return item in self._fieldDict

    def __eq__(self, other):
        """Compares `Records` using their hashes if their hashes are the same then `True` is returned.

        # Parameters

        _other_ : `Record`

        > Another `Record` to be compared against

        # Returns

        `bool`

        > If the `records` are the same then `True` is returned
        """
        if not isinstance(other, type(self)):
            return NotImplemented
        else:
            return self.__hash__() == other.__hash__()

    #Other niceties

    def __str__(self):
        """Makes a string with the title of the file as given by self.title, if there is not one it returns "Untitled record"

        # Returns

        `str`

        > The title of the `Record`
        """
        #This is instead of than redefining for the subclasses
        return "{}({})".format(type(self).__name__, self.get('title', self.id))

    def __repr__(self):
        """Makes a string with the id of the file and its type

        # Returns

        `str`

        > The representation of the `Record`
        """
        if self.bad:
            return "<metaknowledge.{} object BAD>".format(type(self).__name__)
        else:
            return "<metaknowledge.{} object {}>".format(type(self).__name__, self.id)

    def __bytes__(self):
        """Returns the binary form of the original"""
        if self.bad:
            raise BadRecord("This record cannot be converted to binary as the input was malformed.\nThe original line number (if any) is: {} and the original file is: '{}'".format(self._sourceLine, self._sourceFile))
        else:
            strLst = []
            for tag in self._fieldDict.keys():
                for i, value in enumerate(self._fieldDict[tag]):
                    if i == 0:
                        strLst.append(tag + ' ')
                    else:
                        strLst.append('   ')
                    strLst.append(value + '\n')
            strLst.append("ER\n")
            return bytes(''.join(strLst), self.encoding())

    def __getstate__(self):
        #Makes a slightly smaller object than __dict__
        d = self.__dict__.copy()
        #Make copy.copy() produce a shallow copy
        d['_fieldDict'] = d['_fieldDict'].copy()
        return d

    def __setstate__(self, state):
        self.__dict__ = state

    def copy(self):
        """Correctly copies the `Record`

        # Returns

        `Record`

        > A completely decoupled copy of the original
        """
        c = copy.copy(self)
        c._fieldDict = c._fieldDict.copy()
        return c

    #Making these immutable

    @property
    def id(self):
        return self._id

    @property
    def sourceFile(self):
        return self._sourceFile

    @property
    def sourceLine(self):
        return self._sourceLine

class ExtendedRecord(Record, metaclass = abc.ABCMeta):
    """A subclass of `Record` that adds processing to the dictionary. It also cannot be use directly and must be subclassed.

    The `ExtendedRecord` class is a extension of `Record` that is intended for use with the records on scientific papers provided by different organizations such as WOS or Pubmed. The 5 abstract (virtual) methods must be defined for each subclass and define how the data in the different fields is processed and how the record can be rewritten to a file.

    # Processing fields

    When an `ExtendedRecord` is created a dictionary, _fieldDict_, must be provided this contains the raw data from the file reader, usually as lists of strings. `tagProcessingFunc` is a `staticmethod` function that takes in a tag string an returns another function to process it.

    Each tag may also be given a second name, as usually what the they are called in the raw data are not very easy to understand (e.g. `'SO'` is the journal name for WOs records). The mapping from the raw tag (`'SO'`) to the human friendly string (`'journal'`)  is done with the `getAltName` `staticmethod`. `getAltName` takes in a tag string and returns either `None` or the other name for that string. Note, `getAltName` must go both directions `WOSRecord.getAltName(WOSRecord.getAltName('SO')) == 'SO'`.

    The last method for processing entries is `specialFuncs` The following are the special keys for `ExtendedRecords`. These must be the alternate names of tags or strings accepted by the `specialFuncs` method.

    + `'authorsFull'`
    + `'keywords'`
    + `'grants'`
    + `'j9'`
    + `'authorsShort'`
    + `'volume'`
    + `'selfCitation'`
    + `'citations'`
    + `'address'`
    + `'abstract'`
    + `'title'`
    + `'month'`
    + `'year'`
    + `'journal'`
    + `'beginningPage'`
    + `'DOI'`

    `specialFuncs` when given one of these must raise a `KeyError` or return an object of the same type as that returned by the `MedlineRecord` or `WOSRecord`. e.g. `'title'` would return a string giving the title of the record.

    For an example of how this works lets first look at the `'SO'` tag on a `WOSRecord` accessed with the alternate name `'journal'`.

        t = R['journal']

    First the private dictionary `_computedFields` is checked for the key `'title'`, which will fail if this is the first time `'journal'` or `'SO'` has been requested, after this the results will be added to the dictionary to speed up future requests.

    Then the _fieldDict_ will be checked for the key and when that fails the key will go through `getAltName` and be checked again. If the record had a journal entry this will succeed and the raw data will be given to the `tagProcessingFunc` using the same key as _fieldDict_, in this case `SO`.

    The results will then be written to `_computedFields` and returned.

    If the requested key was instead `'grants'` (`g = R['grants']`)the both lookups to _fieldDict_ would have failed and the string `'grants'` would have been given to `specialFuncs` which would return a list of all the grants in the `WOSRecord` (this is always `[]` as WOS does not provided grant information).

    What if the key were not present anywhere? Then the `specialFuncs` should raise a `KeyError` which will be caught then re-raised like a dictionary would with an invalid key look up.

    # File Handling fields

    The two other required methods `encoding` and `writeRecord` define how the records can be rewritten to a file. `encoding` is should return a string giving the encoding python would use, e.g. `'utf-8'` or `'latin-1'`. This is the same encoding that the files written by `writeRecord` should have, `writeRecord` when called should write the original record to the provided open file, _infile_. The opening, closing, header and footer of the file will be handled by `RecordCollection`'s `writeFile` function which should me modified accordingly. If the order of the fields in a record is important you can use a [collections.OrderedDict](https://docs.python.org/3/library/collections.html#collections.OrderedDict) for _fieldDict_.

    # \_\_Init\_\_

    The `__init__` of `ExtendedRecord` takes the same arguments as [Record](./Record.html#metaknowledge.Record)
    """
    #Overwriting the Record attribute
    _documented = ['encoding', 'getAltName', 'specialFuncs', 'tagProcessingFunc', 'writeRecord']

    def __init__(self, fieldDict, idValue, bad, error, sFile = "", sLine = 0):
        """Base constructor for Records

        _fieldDict_ : is the unpared entry dict with tags as keys and their lines as a list of strings

        _idValue_ : is the unique ID of the Record, e.g. the WOS number

        _titleKey_ : is the tag giving the title of the Record, e.g. the WOS tag is `'TI'`

        _bad_ : is the bool to flag the Record as having encountered an errror

        _error_ : is the error that bad indicates

        _sFile_ : is the name of the source file

        _sLine_ : is the line number of the start of the Record entry

        _altNames_ : is a dict that maps the names of tags to an alternative name, i.e. the long names dict. It **must** be bidirectional: map long to short and short to long

        _proccessingFuncs_ : is a dict of functions to proccess the tags. It has the short names as keys and their proccessing fucntions as values. Missing tags will result in the unparsed value to be returned.

        The Records inheting from this must implement, calling the implementations in Record with super() will not cause errors:
        + writeRecord
        + tagProcessingFunc
        + encoding
        + titleTag
        + getAltName
        """
        Record.__init__(self, fieldDict, idValue, bad, error, sFile = sFile, sLine = sLine)

        #Memoizing stuff
        self._computedFields = {}

    def __contains__(self, item):
        """Checks if the tag _item_ is in the Record"""
        #Check all the dicts
        #They are ordered assuming people are checking fresh Records not ones with full _computedFields
        if item in self._fieldDict or self.getAltName(item) in self._fieldDict or item in self._computedFields:
            return True
        else:
            try:
                computedVal = self.specialFuncs(item)
            except KeyError:
                return False
            self._computedFields[item] = computedVal
            alt = self.getAltName(item)
            if alt is not None:
                self._computedFields[alt] = computedVal
            return True


    def __getitem__(self, key):
        """Processes the tag requested with _key_ and memoize it.

        Allows long names, but will still raise a KeyError if the tag is missing, regardless of name used.
        """
        try:
            return self._computedFields[key]
        except KeyError:
            if isinstance(key, str):
                if key in self._fieldDict:
                    computedVal = self.tagProcessingFunc(key)(self._fieldDict[key])
                elif self.getAltName(key) in self._fieldDict:
                    key = self.getAltName(key)
                    computedVal = self.tagProcessingFunc(key)(self._fieldDict[key])
                else:
                    try:
                        computedVal = self.specialFuncs(key)
                    except KeyError:
                        raise KeyError("'{}' could not be found in the Record".format(key)) from BaseException
                #Both refer to the same object, computedVal
                self._computedFields[key] = computedVal
                alt = self.getAltName(key)
                if alt is not None:
                    self._computedFields[alt] = computedVal
                return computedVal
            else:
                raise TypeError("Keys to Records must be strings they cannot be of the type '{}'.".format(type(key).__name__)) from BaseException

    #Extra options added to the defaults to make access to raw data easier
    def get(self, tag, default = None, raw = False):
        """Allows access to the raw values or is an Exception safe wrapper to `__getitem__`.

        # Parameters

        _tag_ : `str`

        > The requested tag

        _default_ : `optional [Object]`

        > Default `None`, the object returned when _tag_ is not found

        _raw_ : `optional [bool]`

        > Default `False`, if `True` the unprocessed value of _tag_ is returned

        # Returns

        `Object`

        > The processed value of _tag_ or _default_
        """
        if raw:
            if tag in self._fieldDict:
                return self._fieldDict[tag]
            elif self.getAltName(tag) in self._fieldDict:
                return self._fieldDict[self.getAltName(tag)]
            else:
                return default
        else:
            try:
                return self[tag]
            except KeyError:
                return default

    def values(self, raw = False):
        """Like `values` for dicts but with a `raw` option

        # Parameters

        _raw_ : `optional [bool]`

        > Default `False`, if `True` the `ValuesView` contains the raw values

        # Returns

        `ValuesView`

        > The values of the record
        """
        if raw:
            return self._fieldDict.values()
        else:
            return collections.abc.Mapping.values(self)

    #Keys given by the mixin

    def items(self, raw = False):
        """Like `items` for dicts but with a `raw` option

        # Parameters

        _raw_ : `optional [bool]`

        > Default `False`, if `True` the `KeysView` contains the raw values as the values

        # Returns

        `KeysView`

        > The key-value pairs of the record
        """
        if raw:
            return self._fieldDict.items()
        else:
            return collections.abc.Mapping.items(self)

    #these need a slight tweak over Record's thanks to `_computedFields`

    def __getstate__(self):
        #Makes a slightly smaller object than __dict__
        d = self.__dict__.copy()
        d['_computedFields'] = {}
        #Make copy.copy() produce a shallow copy
        d['_fieldDict'] = d['_fieldDict'].copy()
        return d

    #Making the 'virtual' methods

    @abc.abstractmethod
    def writeRecord(self, infile):
        """An `abstractmethod`, writes the record in its original form to _infile_

        # Parameters

        _infile_ : `writable file`

        > The file to be written to
        """
        pass

    @abc.abstractmethod
    def encoding(self):
        """An `abstractmethod`, gives the encoding string of the record.

        # Returns

        `str`

        > The encoding
        """
        return 'utf-8' #Most likely to be the encoding

    @staticmethod
    def getAltName(tag):
        """An `abstractmethod`, gives the alternate name of _tag_ or `None`

        # Parameters

        _tag_ : `str`

        > The requested tag

        # Returns

        `str`

        > The alternate name of _tag_ or `None`
        """
        return None #Default to Null case

    @staticmethod
    @abc.abstractmethod
    def tagProcessingFunc(tag):
        """An `abstractmethod`, gives the function for processing _tag_

        # Parameters

        _tag_ : `optional [str]`

        > The tag in need of processing

        # Returns

        `function`

        > The function to process the raw tag
        """
        #Should not raise an exception
        return lambda x: x

    @abc.abstractmethod
    def specialFuncs(self, key):
        """An `abstractmethod`, process the special tag, _key_ using the whole `Record`

        # Parameters

        _key_ : `str`

        > One of the special tags: `'authorsFull'`, `'keywords'`, `'grants'`, `'j9'`, `'authorsShort'`, `'volume'`, `'selfCitation'`, `'citations'`, `'address'`, `'abstract'`, `'title'`, `'month'`, `'year'`, `'journal'`, `'beginningPage'` and `'DOI'`

        # Returns

        > The processed value of _key_
        """
        raise KeyError("There are no special functions given by default.")

    @property
    def title(self):
        return self.get('title', default = "Untitled record")

    @property
    def authors(self):
        auth = self.get('authorsFull')
        if auth is None:
            return self.get('authorsShort')
        else:
            return auth

    def getCitations(self, field = None, values = None, pandasFriendly = True):
        """Creates a pandas ready dict with each row a different citation and columns containing the original string, year, journal and author's name.

        There are also options to filter the output citations with _field_ and _values_

        # Parameters

        _field_ : `optional str`

        > Default `None`, if given all citations missing the named field will be dropped.

        _values_ : `optional str or list[str]`

        > Default `None`, if _field_ is also given only those citations with one of the strings given in _values_ will be included.

        > e.g. to get only citations from 1990 or 1991: `field = year, values = [1991, 1990]`

        _pandasFriendly_ : `optional bool`

        > Default `True`, if `False` a list of the citations will be returned instead of the more complicated pandas dict

        # Returns

        `dict`

        > A pandas ready dict with all the citations
        """
        retCites = []
        if values is not None:
            if isinstance(values, (str, int, float)) or not isinstance(values, collections.abc.Container):
                values = [values]
        if field is not None:
            for cite in self.get('citations', []):
                try:
                    targetVal = getattr(cite, field)
                    if values is None or targetVal in values:
                        retCites.append(cite)
                except AttributeError:
                    pass
        else:
            retCites = self.get('citations', [])
        if pandasFriendly:
            return _pandasPrep(retCites, False)
        return retCites

    def subDict(self, tags, raw = False):
        """Creates a dict of values of _tags_ from the Record. The tags are the keys and the values are the values. If the tag is missing the value will be `None`.

        # Parameters

        _tags_ : `list[str]`

        > The list of tags requested

        _raw_ : `optional [bool]`

        >default `False` if `True` the retuned values of the dict will be unprocessed

        # Returns

        `dict`

        > A dictionary with the keys _tags_ and the values from the record
        """
        retDict = {}
        for tag in tags:
            retDict[tag] = self.get(tag, raw = raw)
        return retDict

    def createCitation(self, multiCite = False):
        """Creates a citation string, using the same format as other WOS citations, for the [Record](./Record.html#metaknowledge.Record) by reading the relevant special tags (`'year'`, `'J9'`, `'volume'`, `'beginningPage'`, `'DOI'`) and using it to create a [Citation](./Citation.html#metaknowledge.citation.Citation) object.

        # Parameters

        _multiCite_ : `optional [bool]`

        > Default `False`, if `True` a tuple of Citations is returned with each having a different one of the records authors as the author

        # Returns

        `Citation`

        > A [Citation](./Citation.html#metaknowledge.citation.Citation) object containing a citation for the Record.
        """
        #Need to put the import here to avoid circular import issues
        from .citation import Citation
        valsLst = []
        if multiCite:
            auths = []
            for auth in self.get("authorsShort", []):
                auths.append(auth.replace(',', ''))
        else:
            if self.get("authorsShort", False):
                valsLst.append(self['authorsShort'][0].replace(',', ''))
        if self.get("year", False):
            valsLst.append(str(self.get('year')))
        if self.get("j9", False):
            valsLst.append(self.get('j9'))
        elif self.get("title", False):
            #No j9 means its probably book so using the books title/leaving blank
            valsLst.append(self.get('title', ''))
        if self.get("volume", False):
            valsLst.append('V' + str(self.get('volume')))
        if self.get("beginningPage", False):
            valsLst.append('P' + str(self.get('beginningPage')))
        if self.get("DOI", False):
            valsLst.append('DOI ' + self.get('DOI'))
        if multiCite and len(auths) > 0:
            return(tuple((Citation(', '.join([a] + valsLst)) for a in auths)))
        elif multiCite:
            return Citation(', '.join(valsLst)),
        else:
            return Citation(', '.join(valsLst))

    def authGenders(self, countsOnly = False, fractionsMode = False, _countsTuple = False):
        """Creates a dict mapping `'Male'`, `'Female'` and `'Unknown'` to lists of the names of all the authors.

        # Parameters

        _countsOnly_ : `optional bool`

        > Default `False`, if `True` the counts (lengths of the lists) will be given instead of the lists of names

        _fractionsMode_ : `optional bool`

        > Default `False`, if `True` the fraction counts (lengths of the lists divided by the total  number of authors) will be given instead of the lists of names. This supersedes _countsOnly_

        # Returns

        `dict[str:str or int]`

        > The mapping of genders to author's names or counts
        """

        authDict = recordGenders(self)
        if _countsTuple or countsOnly or fractionsMode:
            rawList = list(authDict.values())
            countsList = []
            for k in ('Male','Female','Unknown'):
                countsList.append(rawList.count(k))
            if fractionsMode:
                tot = sum(countsList)
                for i in range(3):
                    countsList.append(countsList.pop(0) / tot)
            if _countsTuple:
                return tuple(countsList)
            else:
                return {'Male' : countsList[0], 'Female' : countsList[1], 'Unknown' : countsList[2]}
        else:
            return authDict

    def bibString(self, maxLength = 1000, WOSMode = False, restrictedOutput = False, niceID = True):
        """Makes a string giving the Record as a bibTex entry. If the Record is of a journal article (`PT J`) the bibtext type is set to `'article'`, otherwise it is set to `'misc'`. The ID of the entry is the WOS number and all the Record's fields are given as entries with their long names.

        **Note** This is not meant to be used directly with LaTeX none of the special characters have been escaped and there are a large number of unnecessary fields provided. _niceID_ and _maxLength_ have been provided to make conversions easier.

        **Note** Record entries that are lists have their values seperated with the string `' and '`

        # Parameters

        _maxLength_ : `optional [int]`

        > default 1000, The max length for a continuous string. Most bibTex implementation only allow string to be up to 1000 characters ([source](https://www.cs.arizona.edu/~collberg/Teaching/07.231/BibTeX/bibtex.html)), this splits them up into substrings then uses the native string concatenation (the `'#'` character) to allow for longer strings

        _WOSMode_ : `optional [bool]`

        > default `False`, if `True` the data produced will be unprocessed and use double curly braces. This is the style WOS produces bib files in and mostly macthes that.

        _restrictedOutput_ : `optional [bool]`

        > default `False`, if `True` the tags output will be limited to tose found in `metaknowledge.commonRecordFields`

        _niceID_ : `optional [bool]`

        > default `True`, if `True` the ID used will be derived from the authors, publishing date and title, if `False` it will be the UT tag

        # Returns

        `str`

        > The bibTex string of the Record
        """
        keyEntries = []
        if self.bad:
            raise BadRecord("This record cannot be converted to a bibtex entry as the input was malformed.\nThe original line number (if any) is: {} and the original file is: '{}'".format(self._sourceLine, self._sourceFile))
        if niceID:
            if self.get('authorsFull'):
                bibID = self['authorsFull'][0].title().replace(' ', '').replace(',', '').replace('.','')
            else:
                bibID = ''
            if self.get('year', False):
                bibID += '-' + str(self['year'])
            if self.get('month', False):
                bibID += '-' + str(self['month'])
            if self.get('title', False):
                tiSorted = sorted(self.get('title').split(' '), key = len)
                bibID += '-' + tiSorted.pop().title()
                while len(bibID) < 35 and len(tiSorted) > 0:
                    bibID += '-' + tiSorted.pop().title() #Title Case
            if len(bibID) < 30:
                bibID += str(self.id)
        elif WOSMode:
            bibID = 'ISI:{}'.format(self.id[4:])
        else:
            bibID = str(self.id)
        keyEntries.append("author = {{{{{}}}}},".format(' and '.join(self.get('authorsFull', ['None']))))
        if restrictedOutput:
            tagsIter = ((k, self[k]) for k in commonRecordFields if k in self)
        else:
            tagsIter = self.items()
        if WOSMode:
            for tag, value in tagsIter:
                if isinstance(value, list):
                    keyEntries.append("{} = {{{{{}}}}},".format(tag,'\n   '.join((str(v) for v in value))))
                else:
                    keyEntries.append("{} = {{{{{}}}}},".format(tag, value))
            s = """@{0}{{ {1},\n{2}\n}}""".format('misc', bibID, '\n'.join(keyEntries))
        else:
            for tag, value in tagsIter:
                keyEntries.append("{} = {},".format(tag, _bibFormatter(value, maxLength)))
            s = """@{0}{{ {1},\n    {2}\n}}""".format('misc', bibID, '\n    '.join(keyEntries))
        return s

def _bibFormatter(s, maxLength):
    """Formats a string, list or number to make it good for a bib file by:
        * if too long splits up the string correctly
        * tries to use the best quoting characters
        * expands lists into ' and ' seperated values, as per spec for authors field
    Note, this does not escape characters. LaTeX may have issues with the output
    Max length splitting derived from https://www.cs.arizona.edu/~collberg/Teaching/07.231/BibTeX/bibtex.html
    """
    if isinstance(s, list):
        s = ' and '.join((str(v) for v in s))
    elif not isinstance(s, str):
        s = str(s)
    if len(s) > maxLength:
        s = s.replace('"', '')
        s = [s[i * maxLength: (i + 1) * maxLength] for i in range(len(s) // maxLength )]
        s = '"{}"'.format('" # "'.join(s))
    elif '"' not in s:
        s = '"{}"'.format(s)
    else:
        s = s.replace('{', '\\{').replace('}', '\\}')
        s = '{{{}}}'.format(s)
    return s

def _pandasPrep(cites, addcounts):
    mainValues = ['year', 'journal', 'author']
    retDict = {'citeString' : []}
    for s in mainValues:
        retDict[s] = []
    if addcounts:
        retDict.update({'num-cites' : [], 'fraction-cites-overall' : [], 'fraction-cites-year' : []})
        countsDict = {}
        totCites = 0
        for cite in cites:
            totCites += 1
            try:
                countsDict[cite] += 1
            except KeyError:
                countsDict[cite] = 1
            try:
                cYear = cite.year
            except AttributeError:
                continue
            else:
                try:
                    countsDict[cYear] += 1
                except KeyError:
                    countsDict[cYear] = 1

    for cite in set(cites):
        retDict['citeString'].append(str(cite))
        for s in mainValues:
            try:
                retDict[s].append(getattr(cite, s))
            except AttributeError:
                retDict[s].append(None)
        if addcounts:
            count = countsDict[cite]
            retDict['num-cites'].append(count)
            retDict['fraction-cites-overall'].append(count / totCites)
            try:
                retDict['fraction-cites-year'].append(count / countsDict[cite.year])
            except AttributeError:
                retDict['fraction-cites-year'].append(None)
    return retDict



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
import unittest
import metaknowledge
import metaknowledge.WOS
import os
import filecmp
import networkx as nx

disableJournChecking = True

class TestRecordCollection(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        metaknowledge.VERBOSE_MODE = False
        cls.RCmain = metaknowledge.RecordCollection("metaknowledge/tests/testFile.isi")
        cls.RCbadmain = metaknowledge.RecordCollection("metaknowledge/tests/badFile.isi")

    def setUp(self):
        self.RC = self.RCmain.copy()
        self.RCbad = self.RCbadmain.copy()

    def test_isCollection(self):
        self.assertIsInstance(self.RC, metaknowledge.RecordCollection)
        self.assertEqual(str(metaknowledge.RecordCollection()), "RecordCollection(Empty)")
        self.assertTrue(self.RC == self.RC)

    def test_fullRead(self):
        RC = metaknowledge.RecordCollection("metaknowledge/tests/")
        self.assertEqual(len(RC), 1032)

    def test_caching(self):
        RC = metaknowledge.RecordCollection("metaknowledge/tests/", cached = True, name = 'testingCache', extension = 'testFile.isi')
        self.assertTrue(os.path.isfile("metaknowledge/tests/tests.[testFile.isi].mkRecordDirCache"))
        accessTime = os.stat("metaknowledge/tests/testFile.isi").st_atime
        RC2 = metaknowledge.RecordCollection("metaknowledge/tests/", cached = True, name = 'testingCache', extension = 'testFile.isi')
        self.assertEqual(accessTime, os.stat("metaknowledge/tests/testFile.isi").st_atime)
        RC.dropBadEntries()
        RC2.dropBadEntries()
        self.assertEqual(RC, RC2)
        os.remove("metaknowledge/tests/tests.[testFile.isi].mkRecordDirCache")

    def test_bad(self):
        self.assertTrue(metaknowledge.RecordCollection('metaknowledge/tests/badFile.isi').bad)
        with self.assertRaises(metaknowledge.mkExceptions.RCTypeError):
            metaknowledge.RecordCollection('metaknowledge/tests/testFile.isi', extension = '.txt')
        self.assertEqual(self.RCbad | self.RC, self.RCbad | self.RC )
        self.assertEqual(len(self.RCbad | self.RCbad), 32)
        self.assertFalse(self.RCbad == self.RC)
        self.assertEqual('/Users/Reid/Documents/Work/NetworksLab/metaknowledge/metaknowledge/tests/badFile.isi', self.RCbad.errors.keys().__iter__().__next__())

    def test_badEntries(self):
        badRecs = self.RC.badEntries()
        self.assertTrue(badRecs <= self.RC)
        self.assertTrue(badRecs.pop().bad)
        self.RC.dropBadEntries()

    def test_dropJourn(self):
        RCcopy = self.RC.copy()
        self.RC.dropNonJournals()
        self.assertEqual(len(self.RC), len(RCcopy) - 2)
        self.RC.dropNonJournals(invert = True)
        self.assertEqual(len(self.RC), 0)
        RCcopy.dropNonJournals(ptVal = 'B')
        self.assertEqual(len(RCcopy), 1)

    def test_repr(self):
        self.assertEqual(repr(self.RC), "<metaknowledge.RecordCollection object testFile>")

    def test_hash(self):
        self.assertNotEqual(hash(self.RC), hash(self.RCbad))
        R = self.RC.pop()
        RC = metaknowledge.RecordCollection([R])
        self.assertEqual(hash(RC), hash(hash(R)))

    def test_contains(self):
        R = self.RC.peek()
        self.assertTrue(R in self.RC)
        R = self.RC.pop()
        self.assertFalse(R in self.RC)

    def test_conID(self):
        R = self.RC.peek()
        self.assertTrue(self.RC.containsID(R.id))
        self.assertFalse(self.RC.containsID('234567654'))

    def test_discard(self):
        R = self.RC.peek()
        l = len(self.RC)
        self.RC.discard(R)
        l2 = len(self.RC)
        self.assertEqual(l, l2 + 1)
        self.RC.discard(R)
        self.assertEqual(l2, len(self.RC))

    def test_pop(self):
        R = self.RC.pop()
        self.assertFalse(R in self.RC)
        self.RC.clear()
        with self.assertRaises(KeyError):
            R = self.RC.pop()

    def test_peek(self):
        R = self.RC.peek()
        self.assertTrue(R in self.RC)
        self.RC.clear()
        R = self.RC.peek()
        self.assertTrue(R is None)

    def test_clear(self):
        R = self.RCbad.peek()
        self.assertTrue(self.RCbad.bad)
        self.RCbad.clear()
        self.assertFalse(self.RCbad.bad)
        self.assertFalse(R in self.RCbad)

    def test_remove(self):
        R = self.RC.peek()
        l = len(self.RC)
        self.RC.remove(R)
        self.assertEqual(l, len(self.RC) + 1)
        with self.assertRaises(KeyError):
            self.RC.remove(R)

    def test_equOps(self):
        l = len(self.RC)
        for i in range(10):
            self.RCbad.pop()
        lb = len(self.RCbad)
        RC = metaknowledge.RecordCollection([])
        RC.bad = True
        RC |= self.RC
        self.assertEqual(self.RC, RC)
        RC -= self.RC
        self.assertNotEqual(self.RC, RC)
        RC ^= self.RC
        self.assertEqual(self.RC, RC)
        RC &= self.RCbad
        self.assertNotEqual(self.RC, RC)

    def test_newOps(self):
        l = len(self.RC)
        for i in range(10):
            self.RCbad.pop()
        lb = len(self.RCbad)
        RC = metaknowledge.RecordCollection([])
        RC.bad = True
        RC3 = self.RC | RC
        self.assertEqual(self.RC, RC3)
        RC4 = RC3 - self.RC
        self.assertNotEqual(self.RC, RC4)
        RC5 = RC4 ^ self.RC
        self.assertEqual(self.RC, RC5)
        RC6 = RC5 & self.RCbad
        self.assertNotEqual(self.RC, RC6)

    def test_opErrors(self):
        with self.assertRaises(TypeError):
            self.RC <= 1
        with self.assertRaises(TypeError):
            self.RC >= 1
        self.assertTrue(self.RC != 1)
        with self.assertRaises(TypeError):
            self.RC >= 1
        with self.assertRaises(TypeError):
            self.RC |= 1
        with self.assertRaises(TypeError):
            self.RC ^= 1
        with self.assertRaises(TypeError):
            self.RC &= 1
        with self.assertRaises(TypeError):
            self.RC -= 1
        with self.assertRaises(TypeError):
            self.RC | 1
        with self.assertRaises(TypeError):
            self.RC ^ 1
        with self.assertRaises(TypeError):
            self.RC & 1
        with self.assertRaises(TypeError):
            self.RC - 1

    def test_addRec(self):
        l = len(self.RC)
        R = self.RC.pop()
        self.assertEqual(len(self.RC), l - 1)
        self.RC.add(R)
        self.assertEqual(len(self.RC), l)
        RC2 = metaknowledge.RecordCollection("metaknowledge/tests/TwoPaper.isi")
        self.RC |= RC2
        self.assertEqual(len(self.RC), l + 2)
        with self.assertRaises(metaknowledge.CollectionTypeError):
            self.RC.add(1)

    def test_bytes(self):
        with self.assertRaises(metaknowledge.BadRecord):
            self.assertIsInstance(bytes(self.RC), bytes)
        self.RC.dropBadEntries()
        self.assertIsInstance(bytes(self.RC), bytes)

    def test_WOS(self):
        self.RC.dropBadEntries()
        R = self.RC.peek()
        l = len(self.RC)
        self.assertTrue(R, self.RC.getID(R.id))
        self.assertEqual(len(self.RC), l)
        self.RC.removeID(R.id)
        self.assertEqual(len(self.RC), l - 1)
        self.RC.getID(self.RC.peek().id)
        self.assertEqual(len(self.RC), l - 1)
        self.assertFalse(self.RC.getID(self.RC.pop().id))
        self.RC.discardID('sdfghjkjhgfdfghj')
        self.RC.discardID('WOS:A1979GV55600001')
        with self.assertRaises(KeyError):
            self.RC.removeID('ghjkljhgfdfghjmh')

    def test_directoryRead(self):
        self.assertEqual(len(metaknowledge.RecordCollection('.')), 0)
        self.assertTrue(metaknowledge.RecordCollection('metaknowledge/tests/') >= self.RC)
        self.assertTrue(metaknowledge.RecordCollection('metaknowledge/tests/', extension= '.txt') <= self.RC)

    def test_contentType(self):
        RC = metaknowledge.RecordCollection('metaknowledge/tests/')
        self.assertEqual(RC._collectedTypes, {'MedlineRecord', 'WOSRecord', 'ProQuestRecord', 'ScopusRecord'})
        self.assertEqual(self.RC._collectedTypes, {'WOSRecord'})

    def test_write(self):
        fileName = 'OnePaper2.isi'
        RC = metaknowledge.RecordCollection('metaknowledge/tests/' + fileName)
        RC.writeFile(fileName + '.tmp')
        RC.writeFile()
        self.assertTrue(filecmp.cmp('metaknowledge/tests/' + fileName, fileName + '.tmp'))
        self.assertTrue(filecmp.cmp('metaknowledge/tests/' + fileName, RC.name + '.txt'))
        os.remove(fileName + '.tmp')
        os.remove(RC.name + '.txt')

    def test_writeCSV(self):
        filename = "test_writeCSV_temporaryFile.csv"
        if os.path.isfile(filename):
            os.remove(filename)
        self.RC.writeCSV(filename, onlyTheseTags=['UT', 'PT', 'TI', 'AF','J9' ,'CR', 'pubMedID'], firstTags = ['CR', 'UT', 'J9', 'citations'], csvDelimiter = '∂', csvQuote='≠', listDelimiter= '«', longNames=True, numAuthors = False)
        self.assertTrue(os.path.isfile(filename))
        self.assertEqual(os.path.getsize(filename), 107396)
        os.remove(filename)
        self.RC.writeCSV(filename)
        self.assertTrue(os.path.isfile(filename))
        self.assertEqual(os.path.getsize(filename), 89272)
        os.remove(filename)
        self.RC.writeCSV(splitByTag = 'PY', onlyTheseTags = ['id', 'title', 'authorsFull', 'citations', 'keywords', 'DOI'])
        yearsSt = set()
        for R in self.RC:
            yearsSt.add(str(R.get('PY', 2012)))
        for year in yearsSt:
            f = open("{}-testFile.csv".format(year))
            self.assertEqual(f.readline(), '"id","TI","AF","CR","ID","DI","num-Authors","num-Male","num-Female","num-Unknown"\n')
            self.assertGreater(len(f.readline()), 1)
            f.close()
            os.remove("{}-testFile.csv".format(year))

    def test_writeBib(self):
        filename = 'testFile.bib'
        if os.path.isfile(filename):
            os.remove(filename)
        self.RC.dropBadEntries()
        self.RC.writeBib(maxStringLength = 100)
        self.assertEqual(os.path.getsize(filename), 100418)
        os.remove(filename)
        self.RC.writeBib(fname = filename, wosMode = True, reducedOutput = True, niceIDs = False)
        self.assertEqual(os.path.getsize(filename), 78163)
        os.remove(filename)

    def test_rpys(self):
        d = self.RC.rpys()
        self.assertIn(17, d['count'])
        d = self.RC.rpys(1990, 2000)
        self.assertEqual(len(d['year']), 11)
        for v in d.values():
            for i in v:
                self.assertIsInstance(i, int)

    def test_CopyrightFinder(self):
        l = self.RC.findProbableCopyright()
        self.assertEqual(len(l), 7)
        l = self.RC.findProbableCopyright()
        self.assertTrue(' (C) 2002 Optical Society of America.' in l)

    def test_NLP(self):
        filename = 'NLP_test.csv'
        full = self.RC.forNLP(filename, removeCopyright = True, extraColumns = ['ID'])
        self.assertEqual(len(full), 7)
        self.assertEqual(len(full['id']), 33)
        self.assertEqual(full['keywords'][0], full['ID'][0])
        self.assertTrue(' (C) 2002 Optical Society of America.' in full['copyright'])
        self.assertEqual(os.path.getsize(filename), 14445)
        os.remove(filename)
        dropping = self.RC.forNLP(filename,removeNumbers = False, dropList = ['a', 'and', 'the', 'is'], stemmer = lambda x: x.title())
        self.assertEqual(len(dropping), 5)
        self.assertEqual(len(dropping['id']), 33)
        self.assertEqual(os.path.getsize(filename), 12901)
        os.remove(filename)

    def test_forBurst(self):
        filename = 'Burst_test.csv'
        full = self.RC.forBurst('keywords', outputFile = filename)
        self.assertEqual(len(full), 2)
        self.assertEqual(len(full['year']), 75)
        self.assertIn('guides', full['word'])
        os.remove(filename)

    def test_genderStats(self):
        stats = self.RC.genderStats()
        self.assertEqual(stats, {'Unknown': 65, 'Male': 6, 'Female': 1})
        stats = self.RC.genderStats(asFractions = True)
        self.assertEqual(stats['Male'], 0.08333333333333333)

    def test_getCitations(self):
        cites = self.RC.getCitations()
        self.assertIn('LAUE MV, 1920, RELATIVITATSTHEORIE, V1, P227', cites['citeString'])

    def test_makeDict(self):
        d = self.RC.makeDict(onlyTheseTags = list(metaknowledge.WOS.tagsAndNameSet), longNames = True)
        self.assertEqual(len(d), 65)
        self.assertEqual(len(d['wosString']), len(self.RC))
        if d['eISSN'][0] == '2155-3165':
            self.assertEqual(d['eISSN'][1], None)
        else:
            self.assertEqual(d['eISSN'][0], None)
        self.assertIsInstance(d['citations'], list)
        d = self.RC.makeDict(longNames = False, raw = True, numAuthors = False)
        self.assertEqual(len(d), 45)
        self.assertEqual(len(d['UT']), len(self.RC))
        self.assertIsInstance(d['CR'], list)

    def test_coCite(self):
        Gdefault = self.RC.networkCoCitation(fullInfo = True)
        Gauths = self.RC.networkCoCitation(nodeType = "author", dropAnon = False, detailedCore = True)
        GauthsNoExtra = self.RC.networkCoCitation(nodeType = "author", nodeInfo = False)
        Gunwei = self.RC.networkCoCitation(nodeType = 'original', weighted = False)
        if not disableJournChecking:
            Gjour = self.RC.networkCoCitation(nodeType = "journal", dropNonJournals = True)
        Gyear = self.RC.networkCoCitation(nodeType = "year", fullInfo = True, count = False)
        Gcore = self.RC.networkCoCitation(detailedCore = ['AF','AU', 'DE', 'ID', 'PY'], coreOnly = True)
        Gexplode = self.RC.networkCoCitation(expandedCore = True, keyWords = 'a')
        Gcr = self.RC.networkCoCitation(addCR = True, coreOnly = True)
        self.assertIsInstance(Gdefault, nx.classes.graph.Graph)
        self.assertLessEqual(len(Gdefault.edges()), len(Gunwei.edges()))
        self.assertLessEqual(len(Gdefault.nodes()), len(Gunwei.nodes()))
        self.assertEqual(len(GauthsNoExtra.edges()), len(Gauths.edges()))
        self.assertEqual(len(GauthsNoExtra.nodes()), len(Gauths.nodes()) - 1 )
        self.assertTrue('weight' in list(Gdefault.edges(data = True))[0][2])
        self.assertTrue('info' in list(Gdefault.nodes(data = True))[0][1])
        self.assertTrue('fullCite' in list(Gdefault.nodes(data = True))[0][1])
        self.assertFalse('weight' in list(Gunwei.edges(data = True))[0][2])
        self.assertEqual(metaknowledge.graphStats(Gdefault, sentenceString = True), "The graph has 493 nodes, 13000 edges, 0 isolates, 22 self loops, a density of 0.107282 and a transitivity of 0.611431")
        self.assertEqual(metaknowledge.graphStats(Gauths, sentenceString = True), "The graph has 321 nodes, 6699 edges, 1 isolates, 68 self loops, a density of 0.131094 and a transitivity of 0.598575")
        self.assertEqual(metaknowledge.graphStats(Gyear, sentenceString = True), "The graph has 91 nodes, 1898 edges, 0 isolates, 55 self loops, a density of 0.47033 and a transitivity of 0.702332")
        if not disableJournChecking:
            self.assertEqual(len(Gjour.nodes()), 85)
            self.assertEqual(len(Gjour.edges()), 1195)
            self.assertTrue('info' in Gjour.nodes(data=True)[0][1])
        self.assertTrue('info' in list(Gyear.nodes(data=True))[0][1])
        self.assertTrue('fullCite' in list(Gyear.nodes(data = True))[0][1])
        self.assertEqual(Gcore.node['Costadebeauregard O, 1975, CAN J PHYS']['info'], 'COSTADEBEAUREGARD O, COSTADEBEAUREGARD O')
        self.assertEqual(metaknowledge.graphStats(Gexplode, sentenceString = True), "The graph has 73 nodes, 366 edges, 0 isolates, 5 self loops, a density of 0.140411 and a transitivity of 0.523179")
        self.assertIn('AUDOIN C, 1976, J PHYS E SCI INSTRUM', Gcr.node['Huard S, 1979, CAN J PHYS']['citations'])

    def test_coAuth(self):
        Gdefault = self.RC.networkCoAuthor()
        if not disableJournChecking:
            Gdetailed = self.RC.networkCoAuthor(count = False, weighted = False, detailedInfo = True, dropNonJournals = True)
        self.assertIsInstance(Gdefault, nx.classes.graph.Graph)
        self.assertEqual(len(Gdefault.nodes()), 45)
        self.assertEqual(len(Gdefault.edges()), 46)
        if not disableJournChecking:
            self.assertEqual(metaknowledge.graphStats(Gdetailed, sentenceString = True), 'The graph has 45 nodes, 46 edges, 9 isolates, 0 self loops, a density of 0.0464646 and a transitivity of 0.822581')

    def test_cite(self):
        Gdefault = self.RC.networkCitation(fullInfo = True, count = False, dropAnon = True)
        Ganon = self.RC.networkCitation(dropAnon = False)
        Gauths = self.RC.networkCitation(nodeType = "author", detailedCore = True, dropAnon = True)
        GauthsNoExtra = self.RC.networkCitation(nodeType = "author", nodeInfo = False, dropAnon = True)
        Gunwei = self.RC.networkCitation(nodeType = 'original', weighted = False)
        if not disableJournChecking:
            Gjour = self.RC.networkCitation(nodeType = "author", dropNonJournals = True, nodeInfo = True, count = False)
        Gyear = self.RC.networkCitation(nodeType = "year", nodeInfo = True)
        Gcore = self.RC.networkCitation(detailedCore = True, coreOnly = False)
        Gexplode = self.RC.networkCitation(expandedCore = True, keyWords = ['b', 'c'])
        self.assertIsInstance(Gdefault, nx.classes.digraph.DiGraph)
        self.assertLessEqual(len(Gdefault.edges()), len(Gunwei.edges()))
        self.assertLessEqual(len(Gdefault.nodes()), len(Gunwei.nodes()))
        self.assertEqual(len(GauthsNoExtra.edges()), len(Gauths.edges()))
        self.assertEqual(len(GauthsNoExtra.nodes()), len(Gauths.nodes()))
        self.assertTrue('weight' in list(Gdefault.edges(data = True))[0][2])
        self.assertTrue('info' in list(Gdefault.nodes(data = True))[0][1])
        self.assertFalse('weight' in list(Gunwei.edges(data = True))[0][2])
        self.assertEqual(metaknowledge.graphStats(Gdefault, sentenceString = True), "The graph has 510 nodes, 816 edges, 1 isolates, 0 self loops, a density of 0.00314342 and a transitivity of 0.00600437")
        self.assertEqual(metaknowledge.graphStats(Ganon, sentenceString = True), "The graph has 511 nodes, 817 edges, 0 isolates, 0 self loops, a density of 0.00313495 and a transitivity of 0.00600437")
        self.assertEqual(metaknowledge.graphStats(Gauths, sentenceString = True), "The graph has 324 nodes, 568 edges, 1 isolates, 15 self loops, a density of 0.00542751 and a transitivity of 0.0210315")
        if not disableJournChecking:
            self.assertEqual(len(Gjour.edges()), 432)
            self.assertTrue('info' in list(Gjour.nodes(data=True))[0][1])
        self.assertTrue('info' in list(Gyear.nodes(data=True))[0][1])
        self.assertEqual(Gcore.node['Gilles H, 2002, OPT LETT']['info'], 'WOS:000177484300017, Gilles H, Simple technique for measuring the Goos-Hanchen effect with polarization modulation and a position-sensitive detector, OPTICS LETTERS, 27, 1421')
        self.assertEqual(metaknowledge.graphStats(Gexplode, sentenceString = True), "The graph has 19 nodes, 29 edges, 0 isolates, 3 self loops, a density of 0.0847953 and a transitivity of 0.132075")

    def test_networkBibCoupling(self):
        G = self.RC.networkBibCoupling()
        self.assertEqual(metaknowledge.graphStats(G, sentenceString = True), 'The graph has 32 nodes, 304 edges, 1 isolates, 0 self loops, a density of 0.612903 and a transitivity of 0.836511')

    def test_coOccurnce(self):
        self.assertEqual(sum(self.RC.cooccurrenceCounts('TI', *tuple(self.RC.tags()))['Longitudinal and transverse effects of nonspecular reflection'].values()), 104)

    def test_nLevel(self):
        G = self.RC.networkMultiLevel(*tuple(self.RC.tags()))
        self.assertEqual(metaknowledge.graphStats(G, sentenceString = True), 'The graph has 1187 nodes, 58731 edges, 0 isolates, 59 self loops, a density of 0.0834803 and a transitivity of 0.493814')

    def test_oneMode(self):
        Gcr  = self.RC.networkOneMode('CR')
        Gcite = self.RC.networkOneMode('citations', nodeCount = False, edgeWeight = False)
        GcoCit = self.RC.networkCoCitation()
        Gtit = self.RC.networkOneMode('title')
        stemFunc = lambda x: x[:-1]
        Gstem = self.RC.networkOneMode('keywords', stemmer = stemFunc)
        self.assertEqual(len(Gcite.edges()), len(Gcr.edges()))
        self.assertEqual(len(Gcite.nodes()), len(Gcr.nodes()))
        self.assertAlmostEqual(len(Gcite.nodes()), len(GcoCit.nodes()), delta = 50)
        self.assertEqual(len(self.RC.networkOneMode('D2').nodes()), 0)
        self.assertEqual(len(Gtit.nodes()), 31)
        self.assertEqual(len(Gtit.edges()), 0)
        self.assertEqual(len(self.RC.networkOneMode('email').edges()), 3)
        self.assertEqual(len(self.RC.networkOneMode('UT').nodes()), len(self.RC) - 1)
        self.assertEqual(metaknowledge.graphStats(Gstem, sentenceString = True), 'The graph has 41 nodes, 142 edges, 2 isolates, 0 self loops, a density of 0.173171 and a transitivity of 0.854015')
        self.assertIsInstance(list(Gstem.nodes())[0], str)
        with self.assertRaises(TypeError):
            G = self.RC.networkOneMode(b'Not a Tag')
            del G

    def test_twoMode(self):
        self.RC.dropBadEntries()
        Gutti = self.RC.networkTwoMode('UT', 'title', directed = True, recordType = False)
        Gafwc = self.RC.networkTwoMode('AF', 'WC', nodeCount = False, edgeWeight = False)
        Gd2em = self.RC.networkTwoMode('D2', 'email')
        Gemd2 = self.RC.networkTwoMode('email', 'D2')
        Gstemm = self.RC.networkTwoMode('title', 'title', stemmerTag1 = lambda x: x[:-1], stemmerTag2 = lambda x: x + 's')
        self.assertIsInstance(Gutti, nx.classes.digraph.DiGraph)
        self.assertIsInstance(Gafwc, nx.classes.graph.Graph)
        self.assertEqual(list(Gutti.edges('WOS:A1979GV55600001'))[0][1][:31], "EXPERIMENTS IN PHENOMENOLOGICAL")
        self.assertEqual(len(Gutti.nodes()), 2 * len(self.RC) - 1)
        with self.assertRaises(metaknowledge.TagError):
            G = self.RC.networkTwoMode('TI', b'not a tag')
            del G
        with self.assertRaises(metaknowledge.TagError):
            G = self.RC.networkTwoMode(b'Not a Tag', 'TI')
            del G
        self.assertTrue(nx.is_isomorphic(Gd2em, Gemd2))
        self.assertEqual(metaknowledge.graphStats(Gstemm, sentenceString = True), 'The graph has 62 nodes, 31 edges, 0 isolates, 0 self loops, a density of 0.0163934 and a transitivity of 0')
        self.assertTrue('Optical properties of nanostructured thin filmss' in Gstemm)

    def test_nMode(self):
        G = self.RC.networkMultiMode(metaknowledge.WOS.tagToFullDict.keys())
        Gstem = self.RC.networkMultiMode(metaknowledge.WOS.tagToFullDict.keys(), stemmer = lambda x : x[0])
        self.assertEqual(metaknowledge.graphStats(G, sentenceString = True), 'The graph has 1186 nodes, 38564 edges, 0 isolates, 56 self loops, a density of 0.0549192 and a transitivity of 0.295384')
        self.assertEqual(metaknowledge.graphStats(Gstem, sentenceString = True), 'The graph has 50 nodes, 997 edges, 0 isolates, 35 self loops, a density of 0.828571 and a transitivity of 0.855834')

    def test_localCiteStats(self):
        d = self.RC.localCiteStats()
        dPan = self.RC.localCiteStats(pandasFriendly = True)
        dYear = self.RC.localCiteStats(keyType = 'year')
        self.assertEqual(d[metaknowledge.Citation("Azzam R. M. A., 1977, ELLIPSOMETRY POLARIZ")], 1)
        self.assertEqual(len(dPan['Citations']),len(d))
        self.assertTrue(dPan['Citations'][0] in d)
        self.assertEqual(dYear[2009], 2)

    def test_localCitesOf(self):
        C = metaknowledge.Citation("COSTADEB.O, 1974, LETT NUOVO CIMENTO, V10, P852")
        self.assertEqual("WOS:A1976CW02200002", self.RC.localCitesOf(C).peek().id)
        self.assertEqual(self.RC.localCitesOf(self.RC.peek().id),
         self.RC.localCitesOf(self.RC.peek().createCitation()))

    def test_citeFilter(self):
        RCmin = self.RC.citeFilter('', reverse = True)
        RCmax = self.RC.citeFilter('')
        RCanon = self.RC.citeFilter('', 'anonymous')
        RC1970 = self.RC.citeFilter(1970, 'year')
        RCno1970 = self.RC.citeFilter(1970, 'year', reverse = True)
        RCMELLER = self.RC.citeFilter('meller', 'author')
        self.assertEqual(len(RCmin), 0)
        self.assertEqual(len(RCmax), len(self.RC))
        self.assertEqual(len(RCanon), 1)
        self.assertEqual(len(RC1970), 15)
        self.assertEqual(len(RC1970) + len(RCno1970), len(self.RC))
        self.assertEqual(len(RCMELLER), 1)
        RCnocite = metaknowledge.RecordCollection('metaknowledge/tests/OnePaperNoCites.isi')
        self.assertEqual(len(RCnocite.citeFilter('')), 0)

    def test_yearDiff(self):
        Gdefault = self.RC.networkCitation()
        Gfull = self.RC.networkCitation(nodeType="full")
        Goriginal = self.RC.networkCitation(nodeType="original")
        # Is yearDiff included as an attribute
        self.assertTrue('yearDiff' in list(Gdefault.edges(data=True))[0][2])
        self.assertTrue('yearDiff' in list(Gfull.edges(data=True))[0][2])
        self.assertTrue('yearDiff' in list(Goriginal.edges(data=True))[0][2])
        # Is yearDiff being calculated correctly?
        self.assertEqual(Gdefault["Costadebo, 1974, CR ACAD SCI A MATH"]["Gordon Jp, 1973, PHYS REV A"]["yearDiff"], 1)
        self.assertEqual(Gfull["Costadebo, 1974, CR ACAD SCI A MATH"]["Gordon Jp, 1973, PHYS REV A"]["yearDiff"], 1)
        self.assertEqual(Goriginal["COWAN JJ, 1977, J OPT SOC AM, V67, P1307, DOI 10.1364/JOSA.67.001307"]["GOOS F, 1947, ANN PHYS-BERLIN, V1, P333"]['yearDiff'], 30)

    def test_glimpse(self):
        #These tests do depend on terminal size
        gBasic = self.RC.glimpse()
        gCompact = self.RC.glimpse(compact = True)
        gEmpty = self.RC.glimpse('AF', 'qwertyhujk')
        self.assertIn('RecordCollection glimpse made at:', gBasic)
        self.assertIn('Top Authors\n', gBasic)
        self.assertIn('1 Gilles, H\n', gBasic)

        self.assertIn('|1 JOURNAL OF THE OPTICA', gCompact)
        self.assertIn('|Columns are ranked by num. of occurrences and are independent of one another++', gCompact)
        self.assertIn('qwertyhujk', gEmpty)



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
"""Two functions based on _matplotlib_ for generating nicer looking graphs

This is the only module that depends on anything besides _networkx_, it depends on [numpy](http://www.numpy.org/), [scipy](https://www.scipy.org/) and [matplotlib](http://matplotlib.org/).
"""

from .plotting import graphDensityContourPlot, quickVisual



#Written by Reid McIlroy-Young for Dr. John McLevey, University of Waterloo 2015
import os
import sys

__version__ = '3.4.1'

commonRecordFields = [
    'year',
    'volume',
    'beginningPage',
    'DOI',
    'address',
    'j9',
    'citations',
    'grants',
    'selfCitation',
    'authorsShort',
    'authorsFull',
    'title',
    'journal',
    'keywords',
    'abstract',
    'id',
]


def isInteractive():
    """
    A basic check of if the program is running in interactive mode
    """
    if sys.stdout.isatty() and os.name != 'nt':
        #Hopefully everything but ms supports '\r'
        try:
            import threading
        except ImportError:
            return False
        else:
            return True
    else:
        return False

VERBOSE_MODE = isInteractive()

FAST_CITES = False
